#!/usr/bin/perl -w
#
# sec version 2.0 - sec.pl
# simple event correlation tool
#
# Copyright (C) 2000-2001 Risto Vaarandi
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#

use Getopt::Long;
use POSIX qw(:errno_h :sys_wait_h SEEK_SET SEEK_END setsid);
use Fcntl;


# read options given in commandline

GetOptions( "input=s" => \$inputfile,
            "input_timeout=i" => \$input_timeout,
            "timeout_script=s" => \$timeout_script,
            "reopen_timeout=i" => \$reopen_timeout,
            "conf=s" => \@conffilepat,
            "separator=s" => \$separator,
            "log=s" => \$logfile,
            "pid=s" => \$pidfile,
            "dump=s" => \$dumpfile,
            "cleantime=i" => \$cleantime,
            "bufsize=i" => \$bufsize,
            "quoting!" => \$quoting,
            "tail!" => \$tail,
            "detach!" => \$detach );


if ( !defined($inputfile)  ||  !scalar(@conffilepat)  ||
     !defined($logfile)  ||  !defined($pidfile) ) {

print STDERR << "USAGE";

Version: 2.0

Usage: $0 -input=<inputfile> -log=<logfile> -pid=<pidfile> 
          -conf=<conffile pattern> -conf=<conffile pattern> ...

Optional flags:
  -input_timeout=<input timeout> 
  -timeout_script=<timeout script>
  -reopen_timeout=<reopen timeout>
  -dump=<dumpfile>
  -cleantime=<clean time>
  -bufsize=<input buffer size>
  -quoting, -noquoting
  -tail, -notail
  -detach, -nodetach

Obsolete flags:
  -separator=<separator>

USAGE

exit(1);

}



########################################
# Default values for optional flags
########################################


# If timeout_script was not specified as a flag or incorrect value was
# specified for input_timeout, set input_timeout to 'undef' regardless 
# of its value on command line

if ( !defined($timeout_script) || 
     (defined($input_timeout)  &&  !($input_timeout > 0)) )  
  { $input_timeout = undef; }


# Default value in seconds for time interval that separates two
# subsequent passes of lists

if (!defined($cleantime) || !($cleantime > 0))  { $cleantime = 5; }


# If incorrect value was specified for reopen timeout, set it to 'undef'

if (defined($reopen_timeout)  &&  !($reopen_timeout > 0))
  { $reopen_timeout = undef; }


# Default location of dump file

if (!defined($dumpfile))  { $dumpfile = "/tmp/sec.dump"; }


# Default size of input buffer

if (!defined($bufsize) || !($bufsize > 0))  { $bufsize = 10; }


# If -quoting and -noquoting are not specified, -noquoting is assumed

if (!defined($quoting))  { $quoting = 0; }


# If -tail and -notail are not specified, -tail is assumed

if (!defined($tail))  { $tail = 1; }


# If -detach and -nodetach are not specified, -nodetach is assumed

if (!defined($detach))  { $detach = 0; }


##################################################################


##### Internal constants #####

$poll_timeout = 1;

$separator = " | ";

$INVALIDVALUE = -1;

$SINGLE = 0;
$SINGLE_W_SUPPRESS = 1;
$SINGLE_W_SCRIPT = 2;
$PAIR = 3;
$PAIR_W_WINDOW = 4;
$SINGLE_W_THRESHOLD = 5;
$SINGLE_W_2_THRESHOLDS = 6;
$SUPPRESS = 7;
$CALENDAR = 8;

$SUBSTR = 0;
$REGEXP = 1;

$DONTCONT = 0;
$TAKENEXT = 1;

$NONE = 0;
$LOGONLY = 1;
$SHELLCOMMAND = 2;
$PIPE = 3;
$CREATECONTEXT = 4;
$DELETECONTEXT = 5;
$SETCONTEXT = 6;
$ADD = 7;
$REPORT = 8;
$EVENT = 9;
$RESET = 10;

$OPERAND = 0;
$NEGATION = 1;
$AND = 2;
$OR = 3;
$EXPRESSION = 4;

$EXPRSYMBOL = "\0";

%config_keywords = (
  type => 1,
  continue => 1,
  ptype => 1,
  pattern => 1,
  context => 1,
  desc => 1,
  action => 1,
  window => 1,
  thresh => 1,
  continue2 => 1,
  ptype2 => 1,
  pattern2 => 1,
  context2 => 1,
  desc2 => 1,
  action2 => 1,
  window2 => 1,
  thresh2 => 1,
  time => 1,
  script => 1
);



###############################################################
# ------------------------- FUNCTIONS -------------------------
###############################################################

##############################
# Functions related to logging
##############################


# Parameters: par1 - name of the logfile
# Action: logfile will be opened. Filehandle of the logfile will be
#         saved to the global filehandle LOGFILE.

sub open_logfile {

  my($logfile) = $_[0];

  if (open(LOGFILE, ">>$logfile")) { 

    select LOGFILE;
    $| = 1;
    select STDOUT;

  } else {

    if (-t STDERR  ||  -f STDERR) 
      { print STDERR "Can't open logfile $logfile\n"; }
    exit(1);

  }

}



# Parameters: par1, par2, ... - strings to be logged
# Action: strings par1, par2, ... will be equipped with timestamp and 
#         written to LOGFILE as a single line. If STDERR is connected to 
#         terminal, message will also be written there. If SIGHUP, SIGABRT 
#         or SIGUSR2 signal has arrived but is not processed yet, strings 
#         par1, par2, ... will be placed to a buffer and will be written 
#         to a logfile at a later time.

sub log_msg {

  my($ltime, $msg);

  $ltime = localtime(time());
  $msg = join(" ", @_);

  if ($refresh || $softrefresh || $openlog) { 
    push @logmsgbuffer, "$ltime: $msg\n"; 
  } else { 
    print LOGFILE "$ltime: $msg\n"; 
  }

  if (-t STDERR)  { print STDERR "$msg\n"; }

}



#######################################################
# Functions related to configuration file(s) processing
#######################################################


# Parameters: par1 - value to be checked
# Action: return 1 if par1 is 0 or positive integer, 0 otherwise
#         ($value must consist of 0-9 characters only for 1 to be returned,
#          no leading or trailing whitespace symbols are permitted)

sub is_uinteger {

  my($value) = $_[0];

  return !($value =~ tr/[0-9]//cd);

}



# Parameters: par1, par2, .. - strings
# Action: All 2-byte substrings in par1, par2, .. that denote special 
#         symbols ("\n", "\t", ..) will be replaced with corresponding
#         special symbols

sub subst_specchar {

  my($pos, $pos2);
  my($string, $specchar);


  foreach $string (@_) {

    $pos2 = 0;
 
    for (;;) {

      $pos = index($string, "\\", $pos2);

      if ($pos == -1)  { last; }

      if ($pos == length($string) - 1)  { chop($string); last; }

      $specchar = substr($string, $pos + 1, 1);

      if ($specchar eq "t")  { $specchar = "\t"; }
      elsif ($specchar eq "n")  { $specchar = "\n"; }
      elsif ($specchar eq "r")  { $specchar = "\r"; }
      elsif ($specchar eq "s")  { $specchar = " "; }

      substr($string, $pos, 2) = $specchar;

      $pos2 = $pos + length($specchar);

    }

  }

}



# Parameters: par1 - expression
#             par2 - reference to an array
# Action: parantheses and their contents will be replaced with special 
#         symbols $EXPRSYMBOL in par 1. The expressions inside parantheses 
#         will be returned in par2. Previous content of the array par2 
#         is erased. If par1 was parsed successfully, the modified par1
#         will be returned, otherwise undef is returned.

sub replace_subexpr {

  my($expression) = $_[0];
  my($expr_ref) = $_[1];
  my($i, $j, $l, $pos);


  @{$expr_ref} = ();

  $i = 0;
  $j = 0;
  $l = length($expression);
  $pos = undef;

  while ($i < $l) {

    # process expression par1 from the start and inspect every symbol, 
    # adding 1 to $j for every '(' and subtracting 1 for every ')'

    if (substr($expression, $i, 1) eq "(")  { ++$j; }
    elsif (substr($expression, $i, 1) eq ")")  { --$j; }

    # After observing first '(' save its position to $pos;
    # after observing its counterpart ')' replace everything
    # from '(' to ')' with $EXPRSYMBOL (including possible nested
    # expressions), and save the content of parentheses;
    # if at some point $j becomes negative, the parantheses must
    # be unbalanced

    if ($j == 1  &&  !defined($pos))  { $pos = $i; }

    elsif ($j == 0  &&  defined($pos)) {

      # take symbols starting from position $pos+1 (next symbol after
      # '(') up to position $i-1 (the symbol before ')'), and save
      # the symbols to array

      push @{$expr_ref}, substr($expression, $pos + 1, $i - $pos - 1);

      # replace both the parantheses and the symbols between them 
      # with $EXPRSYMBOL

      substr($expression, $pos, $i - $pos + 1) = $EXPRSYMBOL;

      # set the variables according to changes in expression

      $i = $pos;
      $l = length($expression);
      $pos = undef;

    }

    elsif ($j < 0)  { return undef; }    # extra ')' was found

    ++$i;

  }

  # if the parsing ended with non-zero $j, the parantheses were unbalanced

  if ($j == 0)  { return $expression; }  else { return undef; }

}



# Parameters: par1 - pattern type (string)
#             par2 - pattern
#             par3 - the name of the configuration file
#             par4 - line number in configuration file
# Action: par1 and par2 will be analyzed and pair of integers
#         (pattern type, line count) will be returned (line count shows
#         how many lines the pattern is designed to match). If errors
#         are found when analyzing par1 and par2, error message about
#         improper line par4 in configuration file will be logged.

sub analyze_pattern {

  my($pattype) = $_[0];
  my($pat) = $_[1];
  my($conffile) = $_[2];
  my($lineno) = $_[3];
  my($lines);


  if ($pattype =~ /^regexp(\d*)$/i) {

    if (length($1))  { $lines = $1; }  else { $lines = 1; }

    if ($lines > $bufsize  ||  $lines < 1) {

      log_msg("Rule in $conffile at line $lineno:", 
              "Invalid linecount $lines in '$pattype'");

      return ($INVALIDVALUE, $INVALIDVALUE);

    }

    eval { "" =~ /$pat/; };

    if ($@) {

      log_msg("Rule in $conffile at line $lineno:", 
              "Invalid regular expression '$pat'");

      return ($INVALIDVALUE, $INVALIDVALUE);

    }

    return ($REGEXP, $lines);

  } elsif ($pattype =~ /^substr(\d*)$/i) {

    if (length($1))  { $lines = $1; }  else { $lines = 1; }

    if ($lines > $bufsize  ||  $lines < 1) {

      log_msg("Rule in $conffile at line $lineno:",
              "Invalid linecount $lines in '$pattype'");

      return ($INVALIDVALUE, $INVALIDVALUE);

    }

    return ($SUBSTR, $lines);

  } else {

    log_msg("Rule in $conffile at line $lineno:", 
            "Invalid pattern type '$pattype'");

    return ($INVALIDVALUE, $INVALIDVALUE);

  }

}



# Parameters: par1 - action
#             par2 - the name of the configuration file
#             par3 - line number in configuration file
#             par4 - rule ID
# Action: par1 will be analyzed and pair of integers
#         (action type, action description) will be returned. If errors
#         are found when analyzing par1, error message about improper 
#         line par3 in configuration file will be logged.

sub analyze_action {

  my($action) = $_[0];
  my($conffile) = $_[1];
  my($lineno) = $_[2];
  my($ruleid) = $_[3];
  my($cmdline, $progname);
  my($sign, $rule);
  my($actionlist, @action);
  my($createafter, $event);
  my($lifetime, $context);


  if ($action =~ /^none$/i)  { return $NONE; }

  elsif ($action =~ /^logonly$/i)  { return $LOGONLY; }

  elsif ($action =~ /^shellcmd\s+(.*\S)/i) { 

    $cmdline = $1;
    $progname = (split(/\s+/, $cmdline))[0];

    if (! -f $progname) {

      log_msg("Rule in $conffile at line $lineno:", 
              "Warning - could not find '$progname'");

    } elsif (! -x $progname) {

      log_msg("Rule in $conffile at line $lineno:", 
              "Warning - '$progname' is not executable");

    }

    return ($SHELLCOMMAND, $cmdline); 

  }

  elsif ($action =~ /^pipe\s+'([^']*)'\s+(.*\S)/i) {

    $event = $1;
    $cmdline = $2;

    if (!length($event))  { $event = "%s"; }

    $progname = (split(/\s+/, $cmdline))[0];

    if (! -f $progname) {

      log_msg("Rule in $conffile at line $lineno:", 
              "Warning - could not find '$progname'");

    } elsif (! -x $progname) {

      log_msg("Rule in $conffile at line $lineno:", 
              "Warning - '$progname' is not executable");

    }

    return ($PIPE, $event, $cmdline); 

  }

  elsif ($action =~ /^create\b\s*(\S*)\s*(\d*)\s*(.*)/i) { 

    $context = $1;
    $lifetime = $2;
    $actionlist = $3;

    # strip outer parantheses if they exist
    if ($actionlist =~ /^\s*\((.*)\)\s*$/)  { $actionlist = $1; }

    if (!length($context))  { $context = "%s"; }
    if (!length($lifetime))  { $lifetime = 0; }

    if (!$lifetime  &&  length($actionlist)) {

      log_msg("Rule in $conffile at line $lineno:",
              "Context '$context' has infinite lifetime,",
              "ignoring actionlist '$actionlist'");

      $actionlist = "";

    }

    if (length($actionlist)) {

      if (!analyze_actionlist($actionlist, \@action,
                              $conffile, $lineno, $ruleid)) { return 0; }

      return ($CREATECONTEXT, $context, $lifetime, [ @action ]);

    }

    return ($CREATECONTEXT, $context, $lifetime, []);

  }

  elsif ($action =~ /^delete\b\s*(.*)/i) { 

    $context = $1;

    if (!length($context))  { $context = "%s"; }

    return ($DELETECONTEXT, $context); 

  }

  elsif ($action =~ /^set\s+(\S+)\s+(\d+)\s*(.*)/i) {

    $context = $1;
    $lifetime = $2;
    $actionlist = $3;

    # strip outer parantheses if they exist
    if ($actionlist =~ /^\s*\((.*)\)\s*$/)  { $actionlist = $1; }

    if (!$lifetime  &&  length($actionlist)) {

      log_msg("Rule in $conffile at line $lineno:",
              "Context '$context' has infinite lifetime,",
              "ignoring actionlist '$actionlist'");

      $actionlist = "";

    }

    if (length($actionlist)) {

      if (!analyze_actionlist($actionlist, \@action,
                              $conffile, $lineno, $ruleid)) { return 0; }

      return ($SETCONTEXT, $context, $lifetime, [ @action ]);

    }

    return ($SETCONTEXT, $context, $lifetime, []);

  }

  elsif ($action =~ /^add\s+(\S+)\s*(.*)/i) {

    $context = $1;
    $event = $2;

    if (!length($event))  { $event = "%s"; }

    return ($ADD, $context, $event); 

  }

  elsif ($action =~ /^report\s+(\S+)\s+(.*\S)/i) {

    $context = $1;
    $cmdline = $2;

    $progname = (split(/\s+/, $cmdline))[0];

    if (! -f $progname) {

      log_msg("Rule in $conffile at line $lineno:", 
              "Warning - could not find '$progname'");

    } elsif (! -x $progname) {

      log_msg("Rule in $conffile at line $lineno:", 
              "Warning - '$progname' is not executable");

    }

    return ($REPORT, $context, $cmdline); 

  }

  elsif ($action =~ /^event\b\s*(\d*)\s*(.*)/i) {

    $createafter = $1;
    $event = $2;

    if (!length($createafter))  { $createafter = 0; }
    if (!length($event))  { $event = "%s"; }

    return ($EVENT, $createafter, $event); 

  }

  elsif ($action =~ /^reset\b\s*([\+-]?)(\d*)\s*(.*)/i) { 

    $sign = $1;
    $rule = $2;
    $event = $3;

    if (length($rule)) {

      if ($sign eq "+") { $rule = $ruleid + $rule; }
      elsif ($sign eq "-") { $rule = $ruleid - $rule; }
      elsif (!$rule) { $rule = $ruleid; } 
      else { --$rule; }

    } else { $rule = ""; }

    if (!length($event))  { $event = "%s"; }

    return ($RESET, $conffile, $rule, $event); 

  }

  log_msg("Rule in $conffile at line $lineno:", "Invalid action '$action'");

  return $INVALIDVALUE;

}



# Parameters: par1 - action list separated by semicolons
#             par2 - reference to an array
#             par3 - the name of the configuration file
#             par4 - line number in configuration file
#             par5 - rule ID
# Action: par1 will be split to parts, every part is analyzed and 
#         pair of integers (action type, action description) will be 
#         saved to @{$par2} for that part. Previous content of the array 
#         is erased. If errors are found when analyzing par1, error 
#         message about improper line par4 in configuration file will 
#         be logged.

sub analyze_actionlist {

  my($actionlist) = $_[0];
  my($arrayref) = $_[1];
  my($conffile) = $_[2];
  my($lineno) = $_[3];
  my($ruleid) = $_[4];
  my(@parts, $part);
  my($actiontype, @action);
  my($newactionlist, @list, $expr);
  my($pos, $l);


  @{$arrayref} = ();

  # replace the actions that are in parantheses with special symbols
  # and save the actions to @list

  $newactionlist = replace_subexpr($actionlist, \@list);

  if (!defined($newactionlist))  { return 0; }

  @parts = split(/\s*;\s*/, $newactionlist);

  $l = length($EXPRSYMBOL);

  foreach $part (@parts) {

    # substitute special symbols with expressions 
    # that were removed previously

    for (;;) {

      $pos = index($part, $EXPRSYMBOL);
      if ($pos == -1)  { last; }

      $expr = shift @list;
      substr($part, $pos, $l) = "(" . $expr . ")";

    }

    ($actiontype, @action) = 
        analyze_action($part, $conffile, $lineno, $ruleid);

    if ($actiontype == $INVALIDVALUE)  { return 0; }

    push @{$arrayref}, $actiontype, @action;

  }

  return 1;

}



# Parameters: par1 - context expression
#             par2 - reference to an array
# Action: par1 will be analyzed and saved to array par2 in reverse
#         polish notation form (it is assumed that par1 does not contain
#         expressions in parantheses). Previous content of the array par2 
#         is erased. If errors are found when analyzing par1, 0 will be 
#         returned as a value, otherwise 1 will be returned.

sub analyze_context_expr {

  my($context) = $_[0];
  my($result) = $_[1];
  my($pos, $op1, $op2);
  my(@side1, @side2);


  $pos = rindex($context, "||");

  if ($pos != -1) {

    $op1 = substr($context, 0, $pos);
    $op2 = substr($context, $pos + 2);

    if (!analyze_context_expr($op1, \@side1))  { return 0; }
    if (!analyze_context_expr($op2, \@side2))  { return 0; }

    @{$result} = ( @side1, @side2, $OR );

    return 1;

  }

  $pos = rindex($context, "&&");

  if ($pos != -1) {

    $op1 = substr($context, 0, $pos);
    $op2 = substr($context, $pos + 2);

    if (!analyze_context_expr($op1, \@side1))  { return 0; }
    if (!analyze_context_expr($op2, \@side2))  { return 0; }

    @{$result} = ( @side1, @side2, $AND );

    return 1;

  }

  $pos = index($context, "!");

  if ($pos != -1) {

    $op1 = substr($context, 0, $pos);
    $op2 = substr($context, $pos + 1);

    if ($op1 !~ /^\s*$/)  { return 0; }
    if (!analyze_context_expr($op2, \@side2))  { return 0; }

    @{$result} = ( @side2, $NEGATION );

    return 1;

  }

  if ($context =~ /^\s*(.*\S)/) {

    @{$result} = ( $OPERAND, $1 );

  } else { return 0; }

}



# Parameters: par1 - context description
#             par2 - reference to an array
# Action: par1 will be analyzed and saved to array par2 in reverse
#         polish notation form. Previous content of the array par2 is erased. 
#         If errors are found when analyzing par1, 0 will be returned as 
#         a value, otherwise 1 will be returned.

sub analyze_context {

  my($context) = $_[0];
  my($result) = $_[1];
  my($newcontext, $i, $j);
  my($subexpr, @expr);


  # replace upper level expressions in parantheses with special symbol
  # and save the expressions to @expr (i.e. !(a && (b || c )) || d 
  # becomes !specialsymbol || d, and "a && (b || c )" is saved to @expr);
  # if context was not parsed successfully, exit

  $newcontext = replace_subexpr($context, \@expr);

  if (!defined($newcontext))  { return 0; }

  # convert the context to reverse polish notation, and if there
  # were no paranthesed subexpressions found in the context during
  # previous step, exit

  if (!analyze_context_expr($newcontext, $result))  { return 0; }

  if ($newcontext eq $context)  { return 1; }

  # If the context contains paranthesed subexpressions, analyze and 
  # convert these expression recursively, attaching the results to 
  # the current context

  $i = 0;
  $j = scalar(@{$result});

  while ($i < $j) {
 
    if ($result->[$i] == $OPERAND) {

      if ($result->[$i+1] eq $EXPRSYMBOL) {
 
        $result->[$i] = $EXPRESSION;
        $result->[$i+1] = [];

        $subexpr = shift @expr;

        if (!analyze_context($subexpr, $result->[$i+1]))  { return 0; }
      
      }

      $i += 2;
 
    }

    else { ++$i; }

  }

  return 1;

}



# Parameters: par1 - list of the time values
#             par2 - minimum possible value for time
#             par3 - maximum possible value for time
#             par4 - offset that must be added to every list value
#             par5 - reference to a hash where every list value is added
# Action: take the list definition and find the time values that belong
#         to the list (list definition is given in crontab-style).
#         After the values have been calculated, add an element to the
#         par5 with the key that equals to the calculated value + offset
#         (if offset is 0, then "2,5-7" becomes 2,5,6,7; if offset is -1,
#         min is 1, and max is 12, then "2,5-7,11-" becomes 1,4,5,6,10,11).
#         Before adding elements to par5, its previous content is erased.
#         If par1 is a list specified incorrectly, return value is 0, 
#         otherwise 1 is returned

sub eval_timelist {

  my($spec) = $_[0];
  my($min) = $_[1];
  my($max) = $_[2];
  my($offset) = $_[3];
  my($ref) = $_[4];
  my(@parts, $part);
  my($pos, $range1, $range2);
  my($i, $j);


  # split time specification into parts (by comma) and look what
  # ranges or individual numbers every part defines

  @parts = split(/,/, $spec);
  if (!scalar(@parts))  { return 0; }

  %{$ref} = ();

  foreach $part (@parts) {

    # if part is empty, skip it and take the next part

    if (!length($part))  { next; }

    # if part equals to '*', assume that it defines the range min..max

    if ($part eq "*") {

      # add offset (this also forces numeric context, so "05" becomes "5")
      # and save values to the hash

      $i = $min + $offset;
      $j = $max + $offset;

      while ($i <= $j)  { $ref->{$i++} = 1; }
      next;

    }

    # if part is not empty and not '*', check if it contains '-'

    $pos = index($part, "-");

    if ($pos == -1) {

      # if part does not contain '-', assume it defines a single number

      if (!is_uinteger($part))  { return 0; }
      if ($part < $min  ||  $part > $max)  { return 0; }

      # add offset (this also forces numeric context, so "05" becomes "5")
      # and save value to the hash

      $part += $offset;
      $ref->{$part} = 1;

    } else {

      # if part does contain '-', assume it defines a range

      $range1 = substr($part, 0, $pos);
      $range2 = substr($part, $pos + 1);

      # if left side of the range is missing, assume minimum for the value;
      # if right side of the range is missing, assume maximum for the value;
      # offset is then added to the left and right side of the range
      # (this also forces numeric context, so "05" becomes "5")

      if (length($range1)) {

        if (!is_uinteger($range1))  { return 0; }
        if ($range1 < $min  ||  $range1 > $max)  { return 0; }

        $i = $range1 + $offset;

      } else { $i = $min + $offset; }

      if (length($range2)) {

        if (!is_uinteger($range2))  { return 0; }
        if ($range2 < $min  ||  $range2 > $max)  { return 0; }

        $j = $range2 + $offset;

      } else { $j = $max + $offset; }

      # save values to the hash

      while ($i <= $j)  { $ref->{$i++} = 1; }

    }

  }

  return 1;

}



# Parameters: par1 - time specification
#             par2..par6 - references to the hashes of minutes, hours, 
#                          days, months and weekdays
#             par7 - the name of the configuration file
#             par8 - line number in configuration file
# Action: par1 will be split to parts, every part is analyzed and 
#         results are saved into hashes par2..par6. 
#         Previous content of the hashes is erased. If errors
#         are found when analyzing par1, 0 is returned, otherwise 1
#         will be return value.

sub analyze_timespec {

  my($timespec) = $_[0];
  my($minref) = $_[1];
  my($hourref) = $_[2];
  my($dayref) = $_[3];
  my($monthref) = $_[4];
  my($wdayref) = $_[5];
  my($conffile) = $_[6];
  my($lineno) = $_[7];
  my(@parts);


  @parts = split(/\s+/, $timespec);
  if (!length($parts[0]))  { shift @parts; }

  if (scalar(@parts) != 5) { 

    log_msg("Rule in $conffile at line $lineno:", 
            "Wrong number of elements in time specification"); 
    return 0; 

  }

  # evaluate minute specification (range 0..59, offset 0)

  if (!eval_timelist($parts[0], 0, 59, 0, $minref)) {

    log_msg("Rule in $conffile at line $lineno:", 
            "Invalid minute specification '$parts[0]'"); 
    return 0;

  }

  # evaluate hour specification (range 0..23, offset 0)

  if (!eval_timelist($parts[1], 0, 23, 0, $hourref)) {

    log_msg("Rule in $conffile at line $lineno:", 
            "Invalid hour specification '$parts[1]'"); 
    return 0;

  }

  # evaluate day specification (range 0..31, offset 0)
  # 0 denotes the last day of a month

  if (!eval_timelist($parts[2], 0, 31, 0, $dayref)) {

    log_msg("Rule in $conffile at line $lineno:", 
            "Invalid day specification '$parts[2]'");
    return 0;

  }

  # evaluate month specification (range 1..12, offset -1)

  if (!eval_timelist($parts[3], 1, 12, -1, $monthref)) {

    log_msg("Rule in $conffile at line $lineno:", 
            "Invalid month specification '$parts[3]'");
    return 0;

  }

  # evaluate weekday specification (range 0..7, offset 0)

  if (!eval_timelist($parts[4], 0, 7, 0, $wdayref)) {

    log_msg("Rule in $conffile at line $lineno:", 
            "Invalid weekday specification '$parts[4]'");
    return 0;

  }

  # if 7 was specified as a weekday, also define 0, 
  # since perl uses only 0 for Sunday

  if (exists($wdayref->{"7"}))  { $wdayref->{"0"} = 1; }

  return 1;

}



# Parameters: par1 - reference to a hash containing the rule
#             par2 - list of required keywords for the rule
#             par3 - the type of the rule
#             par4 - the name of the configuration file
#             par5 - line number in configuration file the rule begins at
# Action: check if all required keywords are present in the rule par1

sub missing_keywords {

  my($ref) = $_[0];
  my($keylist) = $_[1];
  my($type) = $_[2];
  my($conffile) = $_[3];
  my($lineno) = $_[4];
  my($key, $error);

  
  $error = 0;

  foreach $key (@{$keylist}) {

    if (!exists($ref->{$key})) {

      log_msg("Rule in $conffile at line $lineno:", 
              "Keyword '$key' missing (needed for the rule type $type)");
      $error = 1;

    }

  }

  return $error;

}



# Parameters: par1 - reference to a hash containing the rule
#             par2 - name of the configuration file
#             par3 - line number in configuration file the rule begins at
#             par4 - rule ID
# Action: check the rule par1 for correctness and save it to
#         global array $configuration{par2} if it is well-defined

sub check_rule {

  my($ref) = $_[0];
  my($conffile) = $_[1];
  my($lineno) = $_[2];
  my($number) = $_[3];
  my($config, @keywords);
  my($type, $progname);
  my($whatnext, $pattype, $patlines, $pattern, @context, @action);
  my($whatnext2, $pattype2, $patlines2, $pattern2, @context2, @action2);
  my(%minutes, %hours, %days, %months, %weekdays);


  $config = $configuration{$conffile};

  if (!exists($ref->{"type"})) { 

    log_msg("Rule in $conffile at line $lineno:", 
            "Keyword 'type' missing");
    return 0;

  }

  $type = uc($ref->{"type"});

  # ------------------------------------------------------------
  # SINGLE rule
  # ------------------------------------------------------------

  if ($type eq "SINGLE") {

    @keywords = ("ptype", "pattern", "desc", "action");

    if (missing_keywords($ref, \@keywords, $type, 
                         $conffile, $lineno))  { return 0; }

    if (!exists($ref->{"continue"})) { $whatnext = $DONTCONT; }
    elsif (uc($ref->{"continue"}) eq "TAKENEXT") { $whatnext = $TAKENEXT; }
    else { $whatnext = $DONTCONT; }

    ($pattype, $patlines) = analyze_pattern($ref->{"ptype"},
                            $ref->{"pattern"}, $conffile, $lineno);

    if ($pattype == $INVALIDVALUE)  { return 0; }

    $pattern = $ref->{"pattern"};

    if ($pattype == $REGEXP)  { $pattern = qr/$pattern/; }
      else { subst_specchar($pattern); }

    if (!analyze_actionlist($ref->{"action"}, \@action, 
                            $conffile, $lineno, $number)) { 

      log_msg("Rule in $conffile at line $lineno:", 
              "Invalid action list '", $ref->{"action"}, "'");
      return 0; 

    }

    if (exists($ref->{"context"})) { 

      if (!analyze_context($ref->{"context"}, \@context)) { 

        log_msg("Rule in $conffile at line $lineno:", 
                "Invalid context specification '", $ref->{"context"}, "'");
        return 0; 

      } 

    } else { @context = (); }

    $config->[$number] = { "ID" => $number, 
                           "Type" => $SINGLE, 
                           "WhatNext" => $whatnext, 
                           "PatType" => $pattype, 
                           "Pattern" => $pattern, 
                           "PatLines" => $patlines, 
                           "Context" => [ @context ],
                           "Desc" => $ref->{"desc"}, 
                           "Action" => [ @action ],
                           "MatchCount" => 0, 
                           "LineNo" => $lineno };
    return 1;

  }


  # ------------------------------------------------------------
  # SINGLE_W_SCRIPT rule
  # ------------------------------------------------------------

  elsif ($type eq "SINGLEWITHSCRIPT") {

    @keywords = ("ptype", "pattern", "script", "desc", "action");

    if (missing_keywords($ref, \@keywords, $type, 
                         $conffile, $lineno))  { return 0; }

    if (!exists($ref->{"continue"})) { $whatnext = $DONTCONT; }
    elsif (uc($ref->{"continue"}) eq "TAKENEXT") { $whatnext = $TAKENEXT; }
    else { $whatnext = $DONTCONT; }

    ($pattype, $patlines) = analyze_pattern($ref->{"ptype"},
                            $ref->{"pattern"}, $conffile, $lineno);

    if ($pattype == $INVALIDVALUE)  { return 0; }

    $pattern = $ref->{"pattern"};

    if ($pattype == $REGEXP)  { $pattern = qr/$pattern/; }
      else { subst_specchar($pattern); }

    $progname = (split(/\s+/, $ref->{"script"}))[0];

    if (! -f $progname) {

      log_msg("Rule in $conffile at line $lineno:", 
              "Warning - could not find '$progname'");

    } elsif (! -x $progname) {

      log_msg("Rule in $conffile at line $lineno:", 
              "Warning - '$progname' is not executable");

    }

    if (!analyze_actionlist($ref->{"action"}, \@action, 
                            $conffile, $lineno, $number)) {

      log_msg("Rule in $conffile at line $lineno:", 
              "Invalid action list '", $ref->{"action"}, "'");
      return 0; 

    }

    if (exists($ref->{"context"})) { 

      if (!analyze_context($ref->{"context"}, \@context)) { 

        log_msg("Rule in $conffile at line $lineno:", 
                "Invalid context specification '", $ref->{"context"}, "'");
        return 0; 

      } 

    } else { @context = (); }

    $config->[$number] = { "ID" => $number, 
                           "Type" => $SINGLE_W_SCRIPT, 
                           "WhatNext" => $whatnext, 
                           "PatType" => $pattype, 
                           "Pattern" => $pattern, 
                           "PatLines" => $patlines,
                           "Context" => [ @context ],
                           "Script" => $ref->{"script"},
                           "Desc" => $ref->{"desc"}, 
                           "Action" => [ @action ],
                           "MatchCount" => 0, 
                           "LineNo" => $lineno };
    return 1;

  }

  # ------------------------------------------------------------
  # SINGLE_W_SUPPRESS rule
  # ------------------------------------------------------------

  elsif ($type eq "SINGLEWITHSUPPRESS") {

    @keywords = ("ptype", "pattern", "desc", "action", "window");

    if (missing_keywords($ref, \@keywords, $type, 
                         $conffile, $lineno))  { return 0; }

    if (!exists($ref->{"continue"})) { $whatnext = $DONTCONT; }
    elsif (uc($ref->{"continue"}) eq "TAKENEXT") { $whatnext = $TAKENEXT; }
    else { $whatnext = $DONTCONT; }

    ($pattype, $patlines) = analyze_pattern($ref->{"ptype"},
                            $ref->{"pattern"}, $conffile, $lineno);

    if ($pattype == $INVALIDVALUE)  { return 0; }

    $pattern = $ref->{"pattern"};

    if ($pattype == $REGEXP)  { $pattern = qr/$pattern/; }
      else { subst_specchar($pattern); }

    if (!analyze_actionlist($ref->{"action"}, \@action, 
                            $conffile, $lineno, $number)) {

      log_msg("Rule in $conffile at line $lineno:", 
              "Invalid action list '", $ref->{"action"}, "'");
      return 0; 

    }

    if (!is_uinteger($ref->{"window"})  ||  $ref->{"window"} == 0) { 

      log_msg("Rule in $conffile at line $lineno:", 
              "Invalid time window '", $ref->{"window"}, "'");
      return 0;

    }

    if (exists($ref->{"context"})) { 

      if (!analyze_context($ref->{"context"}, \@context)) { 

        log_msg("Rule in $conffile at line $lineno:", 
                "Invalid context specification '", $ref->{"context"}, "'");
        return 0; 

      } 

    } else { @context = (); }

    $config->[$number] = { "ID" => $number, 
                           "Type" => $SINGLE_W_SUPPRESS, 
                           "WhatNext" => $whatnext, 
                           "PatType" => $pattype, 
                           "Pattern" => $pattern, 
                           "PatLines" => $patlines,
                           "Context" => [ @context ], 
                           "Desc" => $ref->{"desc"}, 
                           "Action" => [ @action ],
                           "Window" => $ref->{"window"},
                           "MatchCount" => 0, 
                           "LineNo" => $lineno };
    return 1;

  }

  # ------------------------------------------------------------
  # PAIR rule
  # ------------------------------------------------------------

  elsif ($type eq "PAIR") {

    @keywords = ("ptype", "pattern", "desc", "action", 
                 "ptype2", "pattern2", "desc2", "action2");

    if (missing_keywords($ref, \@keywords, $type, 
                         $conffile, $lineno))  { return 0; }

    if (!exists($ref->{"continue"})) { $whatnext = $DONTCONT; }
    elsif (uc($ref->{"continue"}) eq "TAKENEXT") { $whatnext = $TAKENEXT; }
    else { $whatnext = $DONTCONT; }

    ($pattype, $patlines) = analyze_pattern($ref->{"ptype"},
                            $ref->{"pattern"}, $conffile, $lineno);

    if ($pattype == $INVALIDVALUE)  { return 0; }

    $pattern = $ref->{"pattern"};

    if ($pattype == $REGEXP)  { $pattern = qr/$pattern/; }
      else { subst_specchar($pattern); }

    if (!analyze_actionlist($ref->{"action"}, \@action, 
                            $conffile, $lineno, $number)) {

      log_msg("Rule in $conffile at line $lineno:", 
              "Invalid action list '", $ref->{"action"}, "'");
      return 0; 

    }

    if (!exists($ref->{"continue2"})) { $whatnext2 = $DONTCONT; }
    elsif (uc($ref->{"continue2"}) eq "TAKENEXT") { $whatnext2 = $TAKENEXT; }
    else { $whatnext2 = $DONTCONT; }

    ($pattype2, $patlines2) = analyze_pattern($ref->{"ptype2"},
                              $ref->{"pattern2"}, $conffile, $lineno);

    if ($pattype2 == $INVALIDVALUE)  { return 0; }

    $pattern2 = $ref->{"pattern2"};

    if ($pattype2 == $SUBSTR)  { subst_specchar($pattern2); }

    if (!analyze_actionlist($ref->{"action2"}, \@action2, 
                            $conffile, $lineno, $number)) {

      log_msg("Rule in $conffile at line $lineno:", 
              "Invalid action list '", $ref->{"action2"}, "'");
      return 0; 

    }

    if (!exists($ref->{"window"})) { $ref->{"window"} = 0; }

    elsif (!is_uinteger($ref->{"window"})) { 

      log_msg("Rule in $conffile at line $lineno:", 
              "Invalid time window '", $ref->{"window"}, "'");
      return 0;

    }

    if (exists($ref->{"context"})) { 

      if (!analyze_context($ref->{"context"}, \@context)) { 

        log_msg("Rule in $conffile at line $lineno:", 
                "Invalid 1st context specification '", $ref->{"context"}, "'");
        return 0; 

      } 

    } else { @context = (); }

    if (exists($ref->{"context2"})) { 

      if (!analyze_context($ref->{"context2"}, \@context2)) { 

        log_msg("Rule in $conffile at line $lineno:", 
                "Invalid 2nd context specification '", $ref->{"context2"}, "'");
        return 0; 

      } 

    } else { @context2 = (); }

    $config->[$number] = { "ID" => $number, 
                           "Type" => $PAIR, 
                           "WhatNext" => $whatnext, 
                           "PatType" => $pattype, 
                           "Pattern" => $pattern, 
                           "PatLines" => $patlines, 
                           "Context" => [ @context ],
                           "Desc" => $ref->{"desc"}, 
                           "Action" => [ @action ],
                           "WhatNext2" => $whatnext2,
                           "PatType2" => $pattype2,
                           "Pattern2" => $pattern2,
                           "PatLines2" => $patlines2,
                           "Context2" => [ @context2 ],
                           "Desc2" => $ref->{"desc2"},
                           "Action2" => [ @action2 ],
                           "Window" => $ref->{"window"},
                           "Events" => {},
                           "MatchCount" => 0, 
                           "LineNo" => $lineno };
    return 1;

  }

  # ------------------------------------------------------------
  # PAIR_W_WINDOW rule
  # ------------------------------------------------------------

  elsif ($type eq "PAIRWITHWINDOW") {

    @keywords = ("ptype", "pattern", "desc", "action", 
                 "ptype2", "pattern2", "desc2", "action2", "window");

    if (missing_keywords($ref, \@keywords, $type, 
                         $conffile, $lineno))  { return 0; }

    if (!exists($ref->{"continue"})) { $whatnext = $DONTCONT; }
    elsif (uc($ref->{"continue"}) eq "TAKENEXT") { $whatnext = $TAKENEXT; }
    else { $whatnext = $DONTCONT; }

    ($pattype, $patlines) = analyze_pattern($ref->{"ptype"},
                            $ref->{"pattern"}, $conffile, $lineno);

    if ($pattype == $INVALIDVALUE)  { return 0; }

    $pattern = $ref->{"pattern"};

    if ($pattype == $REGEXP)  { $pattern = qr/$pattern/; }
      else { subst_specchar($pattern); }

    if (!analyze_actionlist($ref->{"action"}, \@action, 
                            $conffile, $lineno, $number)) {

      log_msg("Rule in $conffile at line $lineno:", 
              "Invalid action list '", $ref->{"action"}, "'");
      return 0; 

    }

    if (!exists($ref->{"continue2"})) { $whatnext2 = $DONTCONT; }
    elsif (uc($ref->{"continue2"}) eq "TAKENEXT") { $whatnext2 = $TAKENEXT; }
    else { $whatnext2 = $DONTCONT; }

    ($pattype2, $patlines2) = analyze_pattern($ref->{"ptype2"},
                              $ref->{"pattern2"}, $conffile, $lineno);

    if ($pattype2 == $INVALIDVALUE)  { return 0; }

    $pattern2 = $ref->{"pattern2"};

    if ($pattype2 == $SUBSTR)  { subst_specchar($pattern2); }

    if (!analyze_actionlist($ref->{"action2"}, \@action2, 
                            $conffile, $lineno, $number)) {

      log_msg("Rule in $conffile at line $lineno:", 
              "Invalid action list '", $ref->{"action2"}, "'");
      return 0; 

    }

    if (!is_uinteger($ref->{"window"})  ||  $ref->{"window"} == 0) { 

      log_msg("Rule in $conffile at line $lineno:", 
              "Invalid time window '", $ref->{"window"}, "'");
      return 0;

    }

    if (exists($ref->{"context"})) { 

      if (!analyze_context($ref->{"context"}, \@context)) { 

        log_msg("Rule in $conffile at line $lineno:", 
                "Invalid 1st context specification '", $ref->{"context"}, "'");
        return 0; 

      } 

    } else { @context = (); }

    if (exists($ref->{"context2"})) { 

      if (!analyze_context($ref->{"context2"}, \@context2)) { 

        log_msg("Rule in $conffile at line $lineno:", 
                "Invalid 2nd context specification '", $ref->{"context2"}, "'");
        return 0; 

      } 

    } else { @context2 = (); }

    $config->[$number] = { "ID" => $number, 
                           "Type" => $PAIR_W_WINDOW, 
                           "WhatNext" => $whatnext, 
                           "PatType" => $pattype, 
                           "Pattern" => $pattern, 
                           "PatLines" => $patlines, 
                           "Context" => [ @context ],
                           "Desc" => $ref->{"desc"}, 
                           "Action" => [ @action ],
                           "WhatNext2" => $whatnext2,
                           "PatType2" => $pattype2,
                           "Pattern2" => $pattern2,
                           "PatLines2" => $patlines2,
                           "Context2" => [ @context2 ],
                           "Desc2" => $ref->{"desc2"},
                           "Action2" => [ @action2 ],
                           "Window" => $ref->{"window"},
                           "Events" => {},
                           "MatchCount" => 0, 
                           "LineNo" => $lineno };
    return 1;

  }

  # ------------------------------------------------------------
  # SINGLE_W_THRESHOLD rule
  # ------------------------------------------------------------

  elsif ($type eq "SINGLEWITHTHRESHOLD") {

    @keywords = ("ptype", "pattern", 
                 "desc", "action", "window", "thresh");

    if (missing_keywords($ref, \@keywords, $type, 
                         $conffile, $lineno))  { return 0; }

    if (!exists($ref->{"continue"})) { $whatnext = $DONTCONT; }
    elsif (uc($ref->{"continue"}) eq "TAKENEXT") { $whatnext = $TAKENEXT; }
    else { $whatnext = $DONTCONT; }

    ($pattype, $patlines) = analyze_pattern($ref->{"ptype"},
                            $ref->{"pattern"}, $conffile, $lineno);

    if ($pattype == $INVALIDVALUE)  { return 0; }

    $pattern = $ref->{"pattern"};

    if ($pattype == $REGEXP)  { $pattern = qr/$pattern/; }
      else { subst_specchar($pattern); }

    if (!analyze_actionlist($ref->{"action"}, \@action, 
                            $conffile, $lineno, $number)) {

      log_msg("Rule in $conffile at line $lineno:", 
              "Invalid action list '", $ref->{"action"}, "'");
      return 0; 

    }

    if (!is_uinteger($ref->{"window"})  ||  $ref->{"window"} == 0) { 

      log_msg("Rule in $conffile at line $lineno:", 
              "Invalid time window '", $ref->{"window"}, "'");
      return 0;

    }

    if (!is_uinteger($ref->{"thresh"})  ||  $ref->{"thresh"} == 0) { 

      log_msg("Rule in $conffile at line $lineno:", 
              "Invalid threshold '", $ref->{"thresh"}, "'");
      return 0;

    }

    if (exists($ref->{"context"})) { 

      if (!analyze_context($ref->{"context"}, \@context)) { 

        log_msg("Rule in $conffile at line $lineno:", 
                "Invalid context specification '", $ref->{"context"}, "'");
        return 0; 

      } 

    } else { @context = (); }

    $config->[$number] = { "ID" => $number, 
                           "Type" => $SINGLE_W_THRESHOLD, 
                           "WhatNext" => $whatnext, 
                           "PatType" => $pattype, 
                           "Pattern" => $pattern, 
                           "PatLines" => $patlines, 
                           "Context" => [ @context ],
                           "Desc" => $ref->{"desc"}, 
                           "Action" => [ @action ],
                           "Window" => $ref->{"window"},
                           "Threshold" => $ref->{"thresh"},
                           "MatchCount" => 0, 
                           "LineNo" => $lineno };
    return 1;

  }

  # ------------------------------------------------------------
  # SINGLE_W_2_THRESHOLDS rule
  # ------------------------------------------------------------

  elsif ($type eq "SINGLEWITH2THRESHOLDS") {

    @keywords = ("ptype", "pattern", 
                 "desc", "action", "window", "thresh",
                 "desc2", "action2", "window2", "thresh2");

    if (missing_keywords($ref, \@keywords, $type, 
                         $conffile, $lineno))  { return 0; }

    if (!exists($ref->{"continue"})) { $whatnext = $DONTCONT; }
    elsif (uc($ref->{"continue"}) eq "TAKENEXT") { $whatnext = $TAKENEXT; }
    else { $whatnext = $DONTCONT; }

    ($pattype, $patlines) = analyze_pattern($ref->{"ptype"},
                            $ref->{"pattern"}, $conffile, $lineno);

    if ($pattype == $INVALIDVALUE)  { return 0; }

    $pattern = $ref->{"pattern"};

    if ($pattype == $REGEXP)  { $pattern = qr/$pattern/; }
      else { subst_specchar($pattern); }

    if (!analyze_actionlist($ref->{"action"}, \@action, 
                            $conffile, $lineno, $number)) {

      log_msg("Rule in $conffile at line $lineno:", 
              "Invalid action list '", $ref->{"action"}, "'");
      return 0; 

    }

    if (!is_uinteger($ref->{"window"})  ||  $ref->{"window"} == 0) { 

      log_msg("Rule in $conffile at line $lineno:", 
              "Invalid 1st time window '", $ref->{"window"}, "'");
      return 0;

    }

    if (!is_uinteger($ref->{"thresh"})  ||  $ref->{"thresh"} == 0) { 

      log_msg("Rule in $conffile at line $lineno:", 
              "Invalid 1st threshold '", $ref->{"thresh"}, "'");
      return 0;

    }

    if (!analyze_actionlist($ref->{"action2"}, \@action2, 
                            $conffile, $lineno, $number)) {

      log_msg("Rule in $conffile at line $lineno:", 
              "Invalid action list '", $ref->{"action2"}, "'");
      return 0; 

    }

    if (!is_uinteger($ref->{"window2"})  ||  $ref->{"window2"} == 0) { 

      log_msg("Rule in $conffile at line $lineno:", 
              "Invalid 2nd time window '", $ref->{"window2"}, "'");
      return 0;

    }

    if (!is_uinteger($ref->{"thresh2"})) { 

      log_msg("Rule in $conffile at line $lineno:", 
              "Invalid 2nd threshold '", $ref->{"thresh2"}, "'");
      return 0;

    }

    if (exists($ref->{"context"})) { 

      if (!analyze_context($ref->{"context"}, \@context)) { 

        log_msg("Rule in $conffile at line $lineno:", 
                "Invalid context specification '", $ref->{"context"}, "'");
        return 0; 

      } 

    } else { @context = (); }

    $config->[$number] = { "ID" => $number, 
                           "Type" => $SINGLE_W_2_THRESHOLDS, 
                           "WhatNext" => $whatnext, 
                           "PatType" => $pattype, 
                           "Pattern" => $pattern, 
                           "PatLines" => $patlines, 
                           "Context" => [ @context ],
                           "Desc" => $ref->{"desc"}, 
                           "Action" => [ @action ],
                           "Window" => $ref->{"window"},
                           "Threshold" => $ref->{"thresh"},
                           "Desc2" => $ref->{"desc2"},
                           "Action2" => [ @action2 ],
                           "Window2" => $ref->{"window2"},
                           "Threshold2" => $ref->{"thresh2"},
                           "MatchCount" => 0, 
                           "LineNo" => $lineno };
    return 1;

  }

  # ------------------------------------------------------------
  # SUPPRESS rule
  # ------------------------------------------------------------

  elsif ($type eq "SUPPRESS") {

    @keywords = ("ptype", "pattern");

    if (missing_keywords($ref, \@keywords, $type, 
                         $conffile, $lineno))  { return 0; }

    ($pattype, $patlines) = analyze_pattern($ref->{"ptype"},
                            $ref->{"pattern"}, $conffile, $lineno);

    if ($pattype == $INVALIDVALUE)  { return 0; }

    $pattern = $ref->{"pattern"};

    if ($pattype == $REGEXP)  { $pattern = qr/$pattern/; }
      else { subst_specchar($pattern); }

    if (exists($ref->{"context"})) { 

      if (!analyze_context($ref->{"context"}, \@context)) { 

        log_msg("Rule in $conffile at line $lineno:", 
                "Invalid context specification '", $ref->{"context"}, "'");
        return 0; 

      } 

    } else { @context = (); }

    $config->[$number] = { "ID" => $number, 
                           "Type" => $SUPPRESS, 
                           "PatType" => $pattype, 
                           "Pattern" => $pattern, 
                           "PatLines" => $patlines, 
                           "Context" => [ @context ],
                           "Desc" => "Suppress: $pattern",
                           "MatchCount" => 0, 
                           "LineNo" => $lineno };
    return 1;

  }

  # ------------------------------------------------------------
  # CALENDAR rule
  # ------------------------------------------------------------

  elsif ($type eq "CALENDAR") {

    @keywords = ("time", "desc", "action");

    if (missing_keywords($ref, \@keywords, $type, 
                         $conffile, $lineno))  { return 0; }

    if (!analyze_timespec($ref->{"time"}, \%minutes, \%hours, \%days, 
                 \%months, \%weekdays, $conffile, $lineno)) { return 0; }

    if (!analyze_actionlist($ref->{"action"}, \@action, 
                            $conffile, $lineno, $number)) {

      log_msg("Rule in $conffile at line $lineno:", 
              "Invalid action list '", $ref->{"action"}, "'");
      return 0; 

    }

    if (exists($ref->{"context"})) { 

      if (!analyze_context($ref->{"context"}, \@context)) { 

        log_msg("Rule in $conffile at line $lineno:", 
                "Invalid context specification '", $ref->{"context"}, "'");
        return 0; 

      } 

    } else { @context = (); }

    $config->[$number] = { "ID" => $number, 
                           "Type" => $CALENDAR,
                           "Minutes" => { %minutes },
                           "Hours" => { %hours },
                           "Days" => { %days },
                           "Months" => { %months },
                           "Weekdays" => { %weekdays },
                           "LastMinute" => 0,
                           "LastHour" => 0,
                           "LastDay" => 0, 
                           "LastMonth" => 0,
                           "LastWeekday" => 0,  
                           "Context" => [ @context ],
                           "Desc" => $ref->{"desc"},
                           "Action" => [ @action ], 
                           "MatchCount" => 0, 
                           "LineNo" => $lineno };

    push @calendar, $config->[$number];

    return 1;

  }

  # ------------------------------------------------------------
  # unknown rule
  # ------------------------------------------------------------
      
  else {

    log_msg("Rule in $conffile at line $lineno:", 
            "Unknown ruletype $type");
    return 0;

  }

}



# Parameters: par1 - name of the configuration file
# Action: read in rules from configuration file par1 and call
#         check_rule() for every rule

sub read_configfile {

  my($conffile) = $_[0];
  my($linebuf, $line, $i, $cont, $rulestart);
  my($keyword, $value);
  my(%rule);


  log_msg("Reading configuration from $conffile...");

  if (open(CONFFILE, "$conffile")) {

    $configuration{$conffile} = [];

    $i = 0;
    $cont = 0;
    %rule = ();
    $rulestart = 1;

    for (;;) {

      # read next line from file

      $linebuf = <CONFFILE>;

      # check if the line belongs to previous line; if it does, form a 
      # single line from them and start the loop again (i.e. we will
      # concatenate lines until we read a line that does not end with '\')

      if (defined($linebuf)) {
 
        chomp($linebuf);

        if ($cont)  { $line .= $linebuf; }  else { $line = $linebuf; }

        # remove whitespaces from line beginnings and ends;
        # if line is all-whitespace, set it to empty string

        if ($line =~ /^\s*(.*\S)/)  { $line = $1; }  else { $line = ""; }

        # check if line ends with '\'; if it does, remove '\', set $cont
        # to 1 and jump at the start of loop to read next line, otherwise 
        # set $cont to 0

        if (substr($line, length($line) - 1) eq '\\') { 

          chop($line);
          $cont = 1;
          next;

        } else { $cont = 0; }

      }

      # if the line constructed during previous loop is empty, starting 
      # with #-symbol, or if we have reached EOF, consider that as the end 
      # of current rule. Check the rule and set $rulestart to the next line. 
      # If we have reached EOF, quit the loop, otherwise take the next line.

      if (!defined($linebuf) || !length($line) 
                             || index($line, '#') == 0) { 

        if (scalar(%rule)) { 

          if (check_rule(\%rule, $conffile, $rulestart, $i))  { ++$i; }
          %rule = (); 

        }

        $rulestart = $. + 1;
 
        if (defined($linebuf))  { next; }  else { last; }

      }

      # split line into keyword and value

      if ($line =~ /^\s*([^=\s]+)\s*=\s*(.*\S)/) {

        $keyword = $1;
        $value = $2;

      } else {

        log_msg("$conffile line $.:", 
                "Line $line does not conform to keyword=value format");
        next;

      }

      # check if the keyword is valid and save it to hash %rule if it is

      if (!exists($config_keywords{$keyword})) {

        log_msg("$conffile line $.:", "Unknown keyword $keyword");
        next;

      }

      $rule{$keyword} = $value;

    }

    close CONFFILE;

    if (!$i) {

      log_msg("No valid rules found in configuration file $conffile");

      delete $configuration{$conffile};

    } else { log_msg("$i rules loaded from $conffile"); }

  } else {

    log_msg("Can't open configuration file $conffile!");

  }

}



# Parameters: -
# Action: evaluate the conffile patterns given in commandline, form the 
#         list of configuration files and save it to global array 
#         @conffiles, and read in rules from the configuration files

sub read_config {

  my($pattern, $conffile);

  # Initialize global arrays %configuration, @calendar, and @conffiles
  # (the keys for %configuration are members of global array @conffiles)
 
  %configuration = ();
  @calendar = ();
  @conffiles = ();

  # Form the list of configuration files, save it to global array
  # @conffiles, and read configuration from the files

  foreach $pattern (@conffilepat)  { push @conffiles, glob($pattern); }

  foreach $conffile (@conffiles)  { read_configfile($conffile); }

}



#####################################################
# Functions related to processing of lists at runtime
#####################################################


# Parameters: par1 - string
#             par2 - string
# Action: all substrings "%s" in string par1 will be replaced 
#         with string par2

sub substitute_s {

  my($msg) = $_[1];
  my($pos, $pos2, $len);


  $pos2 = 0;
  $len = length($msg);

  for (;;) {

    $pos = index($_[0], "%s", $pos2);
    if ($pos == -1)  { last; }

    substr($_[0], $pos, 2) = $msg;
    $pos2 = $pos + $len;

  }

}



# Parameters: par1 - shell command
#             par2 - message string
# Action: all substrings "%s" in string par1 will be replaced with string
#         "par2", and par1 will be executed as a shell command in a child
#         process. After process has been created, subroutine returns the
#         pid of the child process. If process creation failed, undef is
#         returned. After the command has completed, the child process 
#         terminates and returns command's exit code as its own exit value.

sub shell_cmd {

  my($cmd) = $_[0];
  my($msg) = $_[1];
  my($pid, $ret);


  # try to create a child process and return undef, if fork failed;
  # if fork was successful and we are in parent process, return the 
  # pid of the child process

  $pid = fork();

  if (!defined($pid)) { 

    log_msg("Could not fork command '$cmd'");
    return undef; 

  } elsif ($pid) { 

    log_msg("Child $pid created for command '$cmd'");
    return $pid; 

  }

  # if -quoting flag was specified, replace '-symbols with `-symbols
  # in $msg and put $msg between '-symbols

  if ($quoting) { 

    $msg =~ tr/'/`/;   # ' fix emacs parsing
    $msg = "'" . $msg . "'"; 

  }

  substitute_s($cmd, $msg);

  $ret = system("$cmd");

  # if command was started successfully, return its exit value;
  # if child process was unable to start the command, return 1

  if ($ret != -1)  { exit($ret >> 8); }  else { exit(1); }

}



# Parameters: par1 - shell command for reporting
#             par2 - reference to an array
# Action: par1 will be executed as a shell command in a child process, and
#         contents of array par2 is fed to its standard input. 
#         After process has been created, subroutine returns the
#         pid of the child process. If process creation failed, undef is
#         returned. After the command has completed, the child process 
#         terminates and returns command's exit code as its own exit value.

sub pipe_cmd {

  my($cmd) = $_[0];
  my($ref) = $_[1];
  my($pid, $event);


  # try to create a child process and return undef, if fork failed;
  # if fork was successful and we are in parent process, return the 
  # pid of the child process

  $pid = fork();

  if (!defined($pid)) { 

    log_msg("Could not fork command '$cmd'");
    return undef; 

  } elsif ($pid) { 

    log_msg("Child $pid created for command '$cmd'");
    return $pid; 

  }

  # Try to open a pipe to command. If we are unable to fork the command, 
  # exit with 1; if the command dies and we receive SIGPIPE, exit with 1 
  # also; otherwise exit with the exit value of the command

  $SIG{PIPE} = sub { exit(1); };

  if (open(CMDPIPE, "| $cmd")) {

    select CMDPIPE;
    $| = 1;

    foreach $event (@{$ref})  { print CMDPIPE $event, "\n"; }

    close(CMDPIPE);

    exit($? >> 8);

  } else { 

    exit(1); 

  }

}



# Parameters: par1 - reference to a source action list
#             par2 - reference to a destination action list
# Action: action list par1 will be copied to par2

sub copy_actionlist {

  my($src_ref) = $_[0];
  my($dest_ref) = $_[1];
  my($i, $j);


  @{$dest_ref} = ();
  $i = 0;
  $j = scalar(@{$src_ref});

  while ($i < $j) {

    if ($src_ref->[$i] == $NONE) {
 
      push @{$dest_ref}, $NONE;
      ++$i;
 
    }

    elsif ($src_ref->[$i] == $LOGONLY) {

      push @{$dest_ref}, $LOGONLY; 
      ++$i;
 
    }

    elsif ($src_ref->[$i] == $SHELLCOMMAND) {

      push @{$dest_ref}, $SHELLCOMMAND;
      push @{$dest_ref}, $src_ref->[$i+1]; 
      $i += 2;
 
    }

    elsif ($src_ref->[$i] == $PIPE) {

      push @{$dest_ref}, $PIPE;
      push @{$dest_ref}, $src_ref->[$i+1]; 
      push @{$dest_ref}, $src_ref->[$i+2];
      $i += 3;
 
    }

    elsif ($src_ref->[$i] == $CREATECONTEXT) {

      push @{$dest_ref}, $CREATECONTEXT;
      push @{$dest_ref}, $src_ref->[$i+1]; 
      push @{$dest_ref}, $src_ref->[$i+2];
      push @{$dest_ref}, [];

      copy_actionlist($src_ref->[$i+3], $dest_ref->[$i+3]);
      $i += 4;
 
    }

    elsif ($src_ref->[$i] == $DELETECONTEXT) {

      push @{$dest_ref}, $DELETECONTEXT;
      push @{$dest_ref}, $src_ref->[$i+1]; 
      $i += 2;
 
    }

    elsif ($src_ref->[$i] == $SETCONTEXT) {

      push @{$dest_ref}, $SETCONTEXT;
      push @{$dest_ref}, $src_ref->[$i+1]; 
      push @{$dest_ref}, $src_ref->[$i+2];
      push @{$dest_ref}, [];
 
      copy_actionlist($src_ref->[$i+3], $dest_ref->[$i+3]);
      $i += 4;
 
    }

    elsif ($src_ref->[$i] == $ADD) {

      push @{$dest_ref}, $ADD;
      push @{$dest_ref}, $src_ref->[$i+1]; 
      push @{$dest_ref}, $src_ref->[$i+2];
      $i += 3;
 
    }

    elsif ($src_ref->[$i] == $REPORT) {

      push @{$dest_ref}, $REPORT;
      push @{$dest_ref}, $src_ref->[$i+1]; 
      push @{$dest_ref}, $src_ref->[$i+2];
      $i += 3;
 
    }

    elsif ($src_ref->[$i] == $EVENT) {

      push @{$dest_ref}, $EVENT;
      push @{$dest_ref}, $src_ref->[$i+1]; 
      push @{$dest_ref}, $src_ref->[$i+2];
      $i += 3;
 
    }

    elsif ($src_ref->[$i] == $RESET) {

      push @{$dest_ref}, $RESET;
      push @{$dest_ref}, $src_ref->[$i+1]; 
      push @{$dest_ref}, $src_ref->[$i+2];
      push @{$dest_ref}, $src_ref->[$i+3];
      $i += 4;
 
    }

  }

}



# Parameters: par1 - reference to a source context
#             par2 - reference to a destination context
# Action: context par1 will be copied to par2

sub copy_context {

  my($src_ref) = $_[0];
  my($dest_ref) = $_[1];
  my($i, $j);


  @{$dest_ref} = ();
  $i = 0;
  $j = scalar(@{$src_ref});

  while ($i < $j) {

    if ($src_ref->[$i] == $OPERAND) {

      push @{$dest_ref}, $OPERAND;
      push @{$dest_ref}, $src_ref->[$i+1];
      $i += 2;

    } 

    elsif ($src_ref->[$i] == $EXPRESSION) {

      push @{$dest_ref}, $EXPRESSION;
      push @{$dest_ref}, [];

      copy_context($src_ref->[$i+1], $dest_ref->[$i+1]);
      $i += 2;

    }

    else { 

      push @{$dest_ref}, $src_ref->[$i];
      ++$i; 

    }

  }

}



# Parameters: par1 - reference to a list of actions
#             par2 - event description text
# Action: execute actions in a given action list

sub execute_actionlist {

  my($actionlist) = $_[0];
  my($text) = $_[1];
  my($i, $j);
  my($cmdline, $context, $lifetime, $list);
  my($createafter, $conffile, $ruleid, $event);
  my($key, $ref);


  $i = 0;
  $j = scalar(@{$actionlist});

  while ($i < $j) {

    if ($actionlist->[$i] == $NONE) { 

      ++$i; 

    }

    elsif ($actionlist->[$i] == $LOGONLY) { 

      log_msg($text); 
      ++$i;

    }

    elsif ($actionlist->[$i] == $SHELLCOMMAND) {

      $cmdline = $actionlist->[$i+1];

      log_msg("Executing shell command '$cmdline' for event '$text'...");
      shell_cmd($cmdline, $text);

      $i += 2;

    }

    elsif ($actionlist->[$i] == $PIPE) {

      $event = $actionlist->[$i+1];
      $cmdline = $actionlist->[$i+2];

      substitute_s($event, $text);
      substitute_s($cmdline, $text);

      log_msg("Feeding event '$event' to shell command '$cmdline'...");

      pipe_cmd($cmdline, [ $event ]);

      $i += 3;

    }

    elsif ($actionlist->[$i] == $CREATECONTEXT) {

      $context = $actionlist->[$i+1];
      $lifetime = $actionlist->[$i+2];
      $list = $actionlist->[$i+3];

      substitute_s($context, $text);

      if ( exists($context_list{$context})  &&
           $context_list{$context}->{"Window"}  &&
           time() - $context_list{$context}->{"Time"}
                  > $context_list{$context}->{"Window"}  &&
           !exists($context_list{$context}->{"DeleteInProgress"}) ) {
 
        log_msg("Deleting stale context '$context'...");
 
        if (scalar(@{$context_list{$context}->{"Action"}})) {
 
          $context_list{$context}->{"DeleteInProgress"} = 1;

          execute_actionlist($context_list{$context}->{"Action"},
                             $context_list{$context}->{"Desc"});

        }
 
        delete $context_list{$context};
 
      }

      log_msg("Creating context '$context'...");

      $context_list{$context} = { "Time" => time(), 
                                  "Window" => $lifetime, 
                                  "Buffer" => [],
                                  "Action" => [],
                                  "Desc" => $text };

      copy_actionlist($list, $context_list{$context}->{"Action"});

      $i += 4;

    }

    elsif ($actionlist->[$i] == $DELETECONTEXT) {

      $context = $actionlist->[$i+1];

      substitute_s($context, $text);

      log_msg("Deleting context '$context'...");
      delete $context_list{$context};

      $i += 2;

    }

    elsif ($actionlist->[$i] == $SETCONTEXT) {

      $context = $actionlist->[$i+1];
      $lifetime = $actionlist->[$i+2];
      $list = $actionlist->[$i+3];

      substitute_s($context, $text);

      if ( exists($context_list{$context})  && 
           $context_list{$context}->{"Window"}  &&
           time() - $context_list{$context}->{"Time"}
                  > $context_list{$context}->{"Window"}  &&
           !exists($context_list{$context}->{"DeleteInProgress"}) ) {
 
        log_msg("Deleting stale context '$context'...");
 
        if (scalar(@{$context_list{$context}->{"Action"}})) {
 
          $context_list{$context}->{"DeleteInProgress"} = 1;

          execute_actionlist($context_list{$context}->{"Action"},
                             $context_list{$context}->{"Desc"});

        }
 
        delete $context_list{$context};
 
      }

      log_msg("Changing settings for context '$context'...");

      if (exists($context_list{$context})) {

        $context_list{$context}->{"Time"} = time();
        $context_list{$context}->{"Window"} = $lifetime;
        $context_list{$context}->{"Action"} = [];
        $context_list{$context}->{"Desc"} = $text;

        copy_actionlist($list, $context_list{$context}->{"Action"});

      } else {

        log_msg("Context '$context' does not exist, can't change settings");

      }

      $i += 4;

    }

    elsif ($actionlist->[$i] == $ADD) {

      $context = $actionlist->[$i+1];
      $event = $actionlist->[$i+2];

      substitute_s($context, $text);
      substitute_s($event, $text);

      if ( exists($context_list{$context})  && 
           $context_list{$context}->{"Window"}  &&
           time() - $context_list{$context}->{"Time"}
                  > $context_list{$context}->{"Window"}  &&
           !exists($context_list{$context}->{"DeleteInProgress"}) ) {
 
        log_msg("Deleting stale context '$context'...");
 
        if (scalar(@{$context_list{$context}->{"Action"}})) {

          $context_list{$context}->{"DeleteInProgress"} = 1;
 
          execute_actionlist($context_list{$context}->{"Action"},
                             $context_list{$context}->{"Desc"});

        }
 
        delete $context_list{$context};
 
      }

      log_msg("Adding event '$event' to context '$context'...");

      if (!exists($context_list{$context})) { 

        $context_list{$context} = { "Time" => time(), 
                                    "Window" => 0, 
                                    "Buffer" => [],
                                    "Action" => [],
                                    "Desc" => "" };
      }

      push @{$context_list{$context}->{"Buffer"}}, $event;

      $i += 3;

    }

    elsif ($actionlist->[$i] == $REPORT) {

      $context = $actionlist->[$i+1];
      $cmdline = $actionlist->[$i+2];

      substitute_s($context, $text);
      substitute_s($cmdline, $text);

      if ( exists($context_list{$context})  && 
           $context_list{$context}->{"Window"}  &&
           time() - $context_list{$context}->{"Time"}
                  > $context_list{$context}->{"Window"}  &&
           !exists($context_list{$context}->{"DeleteInProgress"}) ) {
 
        log_msg("Deleting stale context '$context'...");
 
        if (scalar(@{$context_list{$context}->{"Action"}})) {
 
          $context_list{$context}->{"DeleteInProgress"} = 1;

          execute_actionlist($context_list{$context}->{"Action"},
                             $context_list{$context}->{"Desc"});

        }
 
        delete $context_list{$context};
 
      }

      log_msg("Reporting the event store of context '$context' through shell command '$cmdline'...");

      if (!exists($context_list{$context})) {

        log_msg("Context '$context' does not exist, can't report");

      } elsif (!scalar(@{$context_list{$context}->{"Buffer"}})) {

        log_msg("Event store of context '$context' is empty, can't report");

      } else {

        pipe_cmd($cmdline, $context_list{$context}->{"Buffer"});

      }

      $i += 3;

    }

    elsif ($actionlist->[$i] == $EVENT) {

      $createafter = $actionlist->[$i+1];
      $event = $actionlist->[$i+2];

      substitute_s($event, $text);

      if (!$createafter) {

        log_msg("Creating event '$event'...");
        push @events, $event;

      } else { 

        push @pending_events, [ time() + $createafter, $event ]; 

      }

      $i += 3;

    }

    elsif ($actionlist->[$i] == $RESET) {

      $conffile = $actionlist->[$i+1];
      $ruleid = $actionlist->[$i+2];
      $event = $actionlist->[$i+3];

      substitute_s($event, $text);

      if (length($ruleid)) {

        $key = gen_key($conffile, $ruleid, $event);
 
        log_msg("Cancelling the correlation operation with key '$key'...");

        $ref = $configuration{$conffile}->[$ruleid];

        if (exists($ref->{"Events"}))  { delete $ref->{"Events"}->{$key}; }

        delete $corr_list{$key};

      } else {

        log_msg("Cancelling all correlation operations started by rules from",
                 $conffile, "to detect composite event '$event'...");

        foreach $ref (@{$configuration{$conffile}}) {

          $key = gen_key($conffile, $ref->{"ID"}, $event);

          if (exists($ref->{"Events"}))
              { delete $ref->{"Events"}->{$key}; }

          delete $corr_list{$key};

        }

      }

      $i += 4;

    }

  }

}



# Parameters: par1 - context
# Action: check if context "par1" is valid at the moment and return 1
#         if it is, otherwise return 0. If context "par1" is found to
#         be stale but is still present in the context list, it will be
#         removed from there

sub valid_context {

  my($context) = $_[0];


  if (exists($context_list{$context})) {

    if (!$context_list{$context}->{"Window"})  { return 1; }

    if (time() - $context_list{$context}->{"Time"}
          <= $context_list{$context}->{"Window"})  { return 1; }

    if (exists($context_list{$context}->{"DeleteInProgress"}))  { return 0; }

    log_msg("Deleting stale context '$context'...");

    if (scalar(@{$context_list{$context}->{"Action"}})) {

      $context_list{$context}->{"DeleteInProgress"} = 1;

      execute_actionlist($context_list{$context}->{"Action"},
                         $context_list{$context}->{"Desc"});

    }

    delete $context_list{$context};

  }

  return 0;

}



# Parameters: par1 - reference to a context formula
# Action: check if context formula par1 is valid at the moment and return 1
#         if it is, otherwise return 0.

sub valid_formula {

  my($ref) = $_[0];
  my($i, $j, $op1, $op2);
  my(@stack);


  $i = 0;
  $j = scalar(@{$ref});
  @stack = ();

  while ($i < $j) {

    if ($ref->[$i] == $EXPRESSION) {

      $op1 = $ref->[$i+1];
      push @stack, valid_formula($op1);

      $i += 2;
 
    }

    elsif ($ref->[$i] == $OPERAND) {

      $op1 = $ref->[$i+1];
      if (valid_context($op1))  { push @stack, 1; } else { push @stack, 0; }

      $i += 2;
    }

    elsif ($ref->[$i] == $NEGATION) {

      $op1 = pop @stack;
      if ($op1) { push @stack, 0; } else { push @stack, 1; }

      ++$i;
    }

    elsif ($ref->[$i] == $AND) {

      $op1 = pop @stack;
      $op2 = pop @stack;

      if ($op1 && $op2)  { push @stack, 1; } else { push @stack, 0; }

      ++$i;
    }

    elsif ($ref->[$i] == $OR) {

      $op1 = pop @stack;
      $op2 = pop @stack;

      if ($op1 || $op2)  { push @stack, 1; } else { push @stack, 0; }

      ++$i;
    }

  }

  return pop @stack;

}



# Parameters: par1 - number of lines that pattern was designed to match
#             par2 - pattern (string type)
# Action: take par1 last lines from input buffer and concatenate them to 
#         form a single string. Check if par2 is a substring in the formed
#         string (both par1 and par2 can contain newlines), and return 1 
#         if it is, otherwise return 0.

sub match_substr {

  my($linecount) = $_[0];
  my($substr) = $_[1];
  my($i, $line);


  $line = "";
  $i = $bufpos - $linecount + 1;

  for (;;) {

    $line .= $input_buffer[$i % $bufsize];

    if ($i == $bufpos)  { return (index($line, $substr) != -1); }

    $line .= "\n";
    ++$i;

  }

}



# Parameters: par1 - number of lines that pattern was designed to match
#             par2 - pattern (regular expression type)
#             par3 - reference to an array, where backreference values 
#                    $1, $2, .. will be saved. First element of an array will 
#                    be $0 that equals to line(s) that were found matching
# Action: take par1 last lines from input buffer and concatenate them to 
#         form a single string. Match the formed string with regular 
#         expression par2, and if par2 contains bracketing constructs,
#         save backreference values $1, $2, .. to array par3. If formed 
#         string matched regular expression, return 1, otherwise return 0

sub match_regexp {

  my($linecount) = $_[0];
  my($regexp) = $_[1];
  my($subst_ref) = $_[2];
  my($i, $line);


  $line = "";
  $i = $bufpos - $linecount + 1;
  @{$subst_ref} = ();

  for (;;) {

    $line .= $input_buffer[$i % $bufsize];

    if ($i == $bufpos)  { last; }

    $line .= "\n";
    ++$i;

  }

  if (@{$subst_ref} = ($line =~ /$regexp/)) { 

    unshift @{$subst_ref}, $line;   # create $0 that equals to $line
    return 1; 

  } else { 

    return 0; 

  }

}



# Parameters: par1 - reference to the array of replacements
#             par2, par3, .. - strings that will go through replacement
#             procedure
#             par n - token that special variables start with
# Action: Strings par2, par3, .. will be searched for special variables
#         (like $0, $1, $2, ..) that will be replaced with 1st, 2nd, .. 
#         element from array par1 

sub subst_string {

  my($subst_ref) = shift @_;
  my($token) = pop @_;
  my($i, $j, $msg);
  my($pos, $pos2, $len, $len2);


  $j = scalar(@{$subst_ref});

  for ($i = 0; $i < $j; ++$i) {

    $len = length($subst_ref->[$i]);
    $len2 = length("$token$i");

    foreach $msg (@_) {

      $pos2 = 0;

      for (;;) {

        $pos = index($msg, "$token$i", $pos2);
        if ($pos == -1)  { last; }

        substr($msg, $pos, $len2) = $subst_ref->[$i];
        $pos2 = $pos + $len;

      }

    }

  }

}



# Parameters: par1 - reference to the array of replacements
#             par2 - reference to a context formula
#             par3 - token that special variables start with
# Action: Context formula par2 will be searched for special variables
#         (like $1, $2, ..) that will be replaced with 1st, 2nd, .. element
#         from array par1 

sub subst_context {

  my($subst_ref) = $_[0];
  my($ref) = $_[1];
  my($token) = $_[2];
  my($i, $j);


  $i = 0;
  $j = scalar(@{$ref});

  while ($i < $j) {

    if ($ref->[$i] == $OPERAND) {

      subst_string($subst_ref, $ref->[$i+1], $token);
      $i += 2;

    } 

    elsif ($ref->[$i] == $EXPRESSION) {

      subst_context($subst_ref, $ref->[$i+1], $token);
      $i += 2;

    }

    else { ++$i; }

  }

}



# Parameters: par1 - reference to the array of replacements
#             par2 - reference to action list
#             par3 - token that special variables start with
# Action: action list par2 will be searched for special variables
#         (like $1, $2, ..) that will be replaced with 1st, 2nd, .. 
#         element from array par1 

sub subst_actionlist {

  my($subst_ref) = $_[0];
  my($actionlist) = $_[1];
  my($token) = $_[2];
  my($i, $j);


  $i = 0;
  $j = scalar(@{$actionlist});

  while ($i < $j) {

    if ($actionlist->[$i] == $NONE) {
 
      ++$i;
 
    }

    elsif ($actionlist->[$i] == $LOGONLY) {
 
      ++$i;
 
    }

    elsif ($actionlist->[$i] == $SHELLCOMMAND) {
 
      subst_string($subst_ref, $actionlist->[$i+1], $token);
      $i += 2;
 
    }

    elsif ($actionlist->[$i] == $PIPE) {

      subst_string($subst_ref, $actionlist->[$i+1], $token);  
      subst_string($subst_ref, $actionlist->[$i+2], $token);
      $i += 3;
 
    }

    elsif ($actionlist->[$i] == $CREATECONTEXT) {
 
      subst_string($subst_ref, $actionlist->[$i+1], $token);
      subst_actionlist($subst_ref, $actionlist->[$i+3], $token);
      $i += 4;
 
    }

    elsif ($actionlist->[$i] == $DELETECONTEXT) {
 
      subst_string($subst_ref, $actionlist->[$i+1], $token);
      $i += 2;
 
    }

    elsif ($actionlist->[$i] == $SETCONTEXT) {

      subst_string($subst_ref, $actionlist->[$i+1], $token);
      subst_actionlist($subst_ref, $actionlist->[$i+3], $token);
      $i += 4;
 
    }

    elsif ($actionlist->[$i] == $ADD) {

      subst_string($subst_ref, $actionlist->[$i+1], $token);  
      subst_string($subst_ref, $actionlist->[$i+2], $token);
      $i += 3;
 
    }

    elsif ($actionlist->[$i] == $REPORT) {

      subst_string($subst_ref, $actionlist->[$i+1], $token);  
      subst_string($subst_ref, $actionlist->[$i+2], $token);
      $i += 3;
 
    }

    elsif ($actionlist->[$i] == $EVENT) {
 
      subst_string($subst_ref, $actionlist->[$i+2], $token);
      $i += 3;
 
    }

    elsif ($actionlist->[$i] == $RESET) {
 
      subst_string($subst_ref, $actionlist->[$i+3], $token);
      $i += 4;
 
    }

  }

}



# Parameters: par1 - reference to the array of replacements
#             par2, par3, .. - regular expressions that will go through 
#             replacement procedure
#             par n - token that special variables start with
# Action: Regular expressions par2, par3, .. will be searched for special 
#         variables (like $1, $2, ..) that will be replaced with 1st, 
#         2nd, .. element from array par1 

sub subst_regexp {

  my($subst_ref) = shift @_;
  my($token) = pop @_;
  my($i, $j, $msg, $quoted);
  my($pos, $pos2, $len, $len2);


  $j = scalar(@{$subst_ref});

  for ($i = 0; $i < $j; ++$i) {

    $quoted = quotemeta($subst_ref->[$i]);
    $len = length($quoted);
    $len2 = length("$token$i");

    foreach $msg (@_) {

      $pos2 = 0;

      for (;;) {

        $pos = index($msg, "$token$i", $pos2);
        if ($pos == -1)  { last; }

        substr($msg, $pos, $len2) = $quoted;
        $pos2 = $pos + $len;

      }

    }

  }

}



# Parameters: par1 - reference to an element from list %corr_list
#             par2 - time
# Action: search event-time list that is associated with element par1,
#         and remove those elements that are obsolete by time par2

sub update_times {

  my($ref) = $_[0];
  my($time) = $_[1];


  while (scalar(@{$ref->{"Times"}})) {

    if ($time - $ref->{"Times"}->[0] <= $ref->{"Window"})  { last; }
    shift @{$ref->{"Times"}};

  }

  if (scalar(@{$ref->{"Times"}})) { 

    $ref->{"Time"} = $ref->{"Times"}->[0];

  } else { $ref->{"Time"} = 0; }

}



# Parameters: par1, par2, .. - strings
# Action: calculate unique key for strings par1, par2, .. that will be
#         used in correlation lists to distinguish between differents events

sub gen_key {

  return join($separator, @_);

}



# Parameters: par1 - name of the configuration file
# Action: search the rules from configuration file par1 and check, if 
#         there is a matching rule for the current content of input buffer.
#         If matching rule is found, new element (that corresponds to
#         an event correlation operation) will be added to the list 
#         %corr_list. Key for new element is calculated by calling gen_key 
#         function:
#           gen_key(file name, rule number, textual description of event)

sub process_rules {

  my($conffile) = $_[0];
  my($key, $ref, $ref2);
  my($time, $match_found, $i);
  my($desc, $pattern2, $desc2);
  my($pid, $script);
  my($below_threshold, $inside_window);
  my(@subst, @context, @context2); 
  my(@action, @action2);


  foreach $ref (@{$configuration{$conffile}}) {

    # skip CALENDAR rule

    if ($ref->{"Type"} == $CALENDAR)  { next; }

    # Check if last N lines of input buffer match the pattern
    # specified by rule (value of N is also specified by rule)
    # If match was found, set $match_found to 1
    # If there are any backreference values available, assign them 
    # to @subst, otherwise leave @subst empty

    if ($ref->{"PatType"} == $REGEXP) {

      $match_found = 
        match_regexp($ref->{"PatLines"}, $ref->{"Pattern"}, \@subst);

    } elsif ($ref->{"PatType"} == $SUBSTR) {

      $match_found = match_substr($ref->{"PatLines"}, $ref->{"Pattern"});
      @subst = ();

    }


    if ($match_found) {

      # Check if rule context is valid at the moment

      if (scalar(@{$ref->{"Context"}})) {

        copy_context($ref->{"Context"}, \@context);

        if (scalar(@subst))  { subst_context(\@subst, \@context, '$'); }

        # if the rule can't be applied in current context and the rule is 
        # of type Pair*, look also for all active correlation operations 
        # associated with the current rule and check if 2nd pattern matches

        if (!valid_formula(\@context)) {

          if ( ($ref->{"Type"} == $PAIR  ||  $ref->{"Type"} == $PAIR_W_WINDOW)  
               &&  scalar(%{$ref->{"Events"}}) ) {

            if (process_rules2($ref)  &&  
                $ref->{"WhatNext2"} == $DONTCONT)  { return 1; }

          }

          next;

        }

      } else { @context = (); }


      # increment the counter that shows the usage of the rule
      # (just for statistical purposes)

      ++$ref->{"MatchCount"};


      # ------------------------------------------------------------
      # SINGLE rule
      # ------------------------------------------------------------

      if ($ref->{"Type"} == $SINGLE) {

        $desc = $ref->{"Desc"};

        copy_actionlist($ref->{"Action"}, \@action);

        if (scalar(@subst)) { 

          subst_string(\@subst, $desc, '$'); 
          subst_actionlist(\@subst, \@action, '$');

        }

        # execute an action and don't save anything to lists

        execute_actionlist(\@action, $desc);

      }

      # ------------------------------------------------------------
      # SINGLE_W_SCRIPT rule
      # ------------------------------------------------------------

      elsif ($ref->{"Type"} == $SINGLE_W_SCRIPT) {

        $desc = $ref->{"Desc"};
        $script = $ref->{"Script"};

        copy_actionlist($ref->{"Action"}, \@action);

        if (scalar(@subst)) { 

          subst_string(\@subst, $desc, $script, '$'); 
          subst_actionlist(\@subst, \@action, '$');

        }

        $pid = shell_cmd($script, "");

        if (defined($pid)) {

          $children{$pid} = { "Desc" => $desc,
                              "Action" => [ @action ] };
        }

      }

      # ------------------------------------------------------------
      # SINGLE_W_SUPPRESS rule
      # ------------------------------------------------------------

      elsif ($ref->{"Type"} == $SINGLE_W_SUPPRESS) {

        $desc = $ref->{"Desc"};

        if (scalar(@subst))  { subst_string(\@subst, $desc, '$'); }

        $key = gen_key($conffile, $ref->{"ID"}, $desc);
        $time = time();

        if (!exists($corr_list{$key})  ||
            $time - $corr_list{$key}->{"Time"} > $ref->{"Window"}) {

          # 1) if this is first event, start the rule OR
          # 2) if we are outside of time window, we can consider previous
          #    application of this rule finished. So this event is again
          #    first one and rule will be started again

          copy_actionlist($ref->{"Action"}, \@action);

          if (scalar(@subst)) { subst_actionlist(\@subst, \@action, '$'); }

          $corr_list{$key} = { "Time" => $time, 
                               "Type" => $ref->{"Type"}, 
                               "File" => $conffile,
                               "ID" => $ref->{"ID"},
                               "Window" => $ref->{"Window"},
                               "Context" => [ @context ],
                               "Desc" => $desc,
                               "Action" => [ @action ] };

          execute_actionlist(\@action, $desc);

        }

      }

      # ------------------------------------------------------------
      # PAIR rule
      # ------------------------------------------------------------

      elsif ($ref->{"Type"} == $PAIR) {

        $desc = $ref->{"Desc"};

        if (scalar(@subst))  { subst_string(\@subst, $desc, '$'); }

        $key = gen_key($conffile, $ref->{"ID"}, $desc);
        $time = time();

        # 1) if this is first event, start the rule OR
        # 2) if we are outside of time window, we can consider previous
        #    application of this rule finished. So this event is again 
        #    first one and rule will be started again

        if ( !exists($corr_list{$key})  ||  ($ref->{"Window"}  &&
             $time - $corr_list{$key}->{"Time"} > $ref->{"Window"}) ) {

          copy_actionlist($ref->{"Action"}, \@action);
          copy_actionlist($ref->{"Action2"}, \@action2);
          copy_context($ref->{"Context2"}, \@context2);

          $pattern2 = $ref->{"Pattern2"};
          $desc2 = $ref->{"Desc2"};

          if (scalar(@subst)) {

            subst_actionlist(\@subst, \@action, '$');

            if ($ref->{"PatType2"} == $REGEXP) { 

              subst_regexp(\@subst, $pattern2, '$'); 
              $pattern2 = qr/$pattern2/;

              subst_string(\@subst, $desc2, '%');
              subst_context(\@subst, \@context2, '%');
              subst_actionlist(\@subst, \@action2, '%');

            } else {

              subst_string(\@subst, $pattern2, $desc2, '$');
              subst_context(\@subst, \@context2, '$');
              subst_actionlist(\@subst, \@action2, '$');

            }

          }

          $corr_list{$key} = { "Time" => $time,
                               "Type" => $ref->{"Type"},
                               "File" => $conffile,
                               "ID" => $ref->{"ID"},
                               "Window" => $ref->{"Window"},
                               "Context" => [ @context ],
                               "Desc" => $desc,
                               "Action" => [ @action ],
                               "Pattern2" => $pattern2, 
                               "Context2" => [ @context2 ],
                               "Desc2" => $desc2,
                               "Action2" => [ @action2 ] };

          $ref->{"Events"}->{$key} = $corr_list{$key};

          execute_actionlist(\@action, $desc);

        }

      }

      # ------------------------------------------------------------
      # PAIR_W_WINDOW rule
      # ------------------------------------------------------------

      elsif ($ref->{"Type"} == $PAIR_W_WINDOW) {

        $desc = $ref->{"Desc"};

        if (scalar(@subst))  { subst_string(\@subst, $desc, '$'); }

        $key = gen_key($conffile, $ref->{"ID"}, $desc);
        $time = time();

        # if we are outside of time window and event is not removed from
        # corr_list, it is still not resolved and action must be executed
        # after that we can consider previous application of the rule finished

        if ( exists($corr_list{$key})  &&
             $time - $corr_list{$key}->{"Time"} > $ref->{"Window"} ) {

          execute_actionlist($corr_list{$key}->{"Action"}, $desc);

          delete $corr_list{$key};
          delete $ref->{"Events"}->{$key};

        }

        # Since this is first event, start the rule

        if (!exists($corr_list{$key})) {

          copy_actionlist($ref->{"Action"}, \@action);
          copy_actionlist($ref->{"Action2"}, \@action2);
          copy_context($ref->{"Context2"}, \@context2);

          $pattern2 = $ref->{"Pattern2"};
          $desc2 = $ref->{"Desc2"};

          if (scalar(@subst)) {

            subst_actionlist(\@subst, \@action, '$');

            if ($ref->{"PatType2"} == $REGEXP) { 

              subst_regexp(\@subst, $pattern2, '$'); 
              $pattern2 = qr/$pattern2/;

              subst_string(\@subst, $desc2, '%');
              subst_context(\@subst, \@context2, '%');
              subst_actionlist(\@subst, \@action2, '%');

            } else { 

              subst_string(\@subst, $pattern2, $desc2, '$'); 
              subst_context(\@subst, \@context2, '$');
              subst_actionlist(\@subst, \@action2, '$');

            }

          }

          $corr_list{$key} = { "Time" => $time, 
                               "Type" => $ref->{"Type"},
                               "File" => $conffile,
                               "ID" => $ref->{"ID"},
                               "Window" => $ref->{"Window"}, 
                               "Context" => [ @context ],
                               "Desc" => $desc,
                               "Action" => [ @action ], 
                               "Pattern2" => $pattern2, 
                               "Context2" => [ @context2 ],
                               "Desc2" => $desc2,
                               "Action2" => [ @action2 ] };

          $ref->{"Events"}->{$key} = $corr_list{$key};

        }

      }

      # ------------------------------------------------------------
      # SINGLE_W_THRESHOLD rule
      # ------------------------------------------------------------ 

      elsif ($ref->{"Type"} == $SINGLE_W_THRESHOLD) {

        $desc = $ref->{"Desc"};

        if (scalar(@subst))  { subst_string(\@subst, $desc, '$'); }

        $key = gen_key($conffile, $ref->{"ID"}, $desc);
        $time = time();

        # Since this is first event, start the rule by creating
        # corr_list element

        if (!exists($corr_list{$key})) {

          copy_actionlist($ref->{"Action"}, \@action);

          if (scalar(@subst)) { subst_actionlist(\@subst, \@action, '$'); }

          $corr_list{$key} = { "Time" => $time, 
                               "Type" => $ref->{"Type"},
                               "File" => $conffile,
                               "ID" => $ref->{"ID"},
                               "Times" => [], 
                               "Window" => $ref->{"Window"},
                               "Context" => [ @context ],
                               "Desc" => $desc,
                               "Action" => [ @action ],
                               "Threshold" => $ref->{"Threshold"} };

        } 

        $ref2 = $corr_list{$key};

        # inside_window - TRUE if we are still in time window
        # below_threshold - TRUE if we were below threshold before this event

        $inside_window = ($time - $ref2->{"Time"} <= $ref->{"Window"});
        $below_threshold = (scalar(@{$ref2->{"Times"}}) < $ref->{"Threshold"});


        if ($inside_window  &&  $below_threshold) {

          # if we are inside time window and below threshold, increase 
          # the counter, and if new value of the counter equals to threshold, 
          # execute an action

          push @{$ref2->{"Times"}}, $time;

          if (scalar(@{$ref2->{"Times"}}) == $ref->{"Threshold"}) {

            execute_actionlist($ref2->{"Action"}, $desc);

          }

        } 

        elsif ($below_threshold) {

          # if we are already outside time window but still below
          # threshold, slide the window forward

          push @{$ref2->{"Times"}}, $time;

          update_times($ref2, $time);

        }

        elsif (!$inside_window) {

          # if we are both outside time window and above threshold
          # then action was executed in the past and this corr_list 
          # element was used to suppress post-action events.
          # So current event is again first one and rule will be restarted

          copy_actionlist($ref->{"Action"}, \@action);

          if (scalar(@subst)) { subst_actionlist(\@subst, \@action, '$'); }

          $corr_list{$key} = { "Time" => $time, 
                               "Type" => $ref->{"Type"},
                               "File" => $conffile,
                               "ID" => $ref->{"ID"},
                               "Times" => [ $time ], 
                               "Window" => $ref->{"Window"},
                               "Context" => [ @context ],
                               "Desc" => $desc,
                               "Action" => [ @action ],
                               "Threshold" => $ref->{"Threshold"} };

          if ($ref->{"Threshold"} == 1) {

            execute_actionlist(\@action, $desc);

          }

        } 

      }

      # ------------------------------------------------------------
      # SINGLE_W_2_THRESHOLDS rule
      # ------------------------------------------------------------ 

      elsif ($ref->{"Type"} == $SINGLE_W_2_THRESHOLDS) {

        $desc = $ref->{"Desc"};

        if (scalar(@subst))  { subst_string(\@subst, $desc, '$'); }

        $key = gen_key($conffile, $ref->{"ID"}, $desc);
        $time = time();

        # Since this is first event, start the rule by creating
        # corr_list element

        if (!exists($corr_list{$key})) {

          copy_actionlist($ref->{"Action"}, \@action);
          copy_actionlist($ref->{"Action2"}, \@action2);

          $desc2 = $ref->{"Desc2"};

          if (scalar(@subst)) { 

            subst_actionlist(\@subst, \@action, '$');
            subst_string(\@subst, $desc2, '$'); 
            subst_actionlist(\@subst, \@action2, '$');

          }

          $corr_list{$key} = { "Time" => $time, 
                               "Type" => $ref->{"Type"},
                               "File" => $conffile,
                               "ID" => $ref->{"ID"},
                               "Times" => [], 
                               "Window" => $ref->{"Window"}, 
                               "Context" => [ @context ],
                               "Desc" => $desc,
                               "Action" => [ @action ],
                               "Threshold" => $ref->{"Threshold"}, 
                               "2ndPass" => 0,
                               "Window2" => $ref->{"Window2"}, 
                               "Threshold2" => $ref->{"Threshold2"}, 
                               "Desc2" => $desc2,
                               "Action2" => [ @action2 ] };

        } 

        $ref2 = $corr_list{$key};


        # ----- if we are still checking 1st threshold...

        if (!$ref2->{"2ndPass"}) {

          # inside_window - TRUE if we are still in time window
          # below_threshold - TRUE if we were below threshold before this event

          $inside_window = ($time - $ref2->{"Time"} <= $ref->{"Window"});
          $below_threshold = (scalar(@{$ref2->{"Times"}}) < $ref->{"Threshold"});


          if ($inside_window) {

            # if we are inside time window, increase the counter, and
            # if new value of the counter equals to threshold, execute
            # an action and start to check 2nd threshold

            push @{$ref2->{"Times"}}, $time;

            if (scalar(@{$ref2->{"Times"}}) == $ref->{"Threshold"}) {

              $ref2->{"Time"} = $time;
              $ref2->{"2ndPass"} = 1;
              $ref2->{"Times"} = [];

              execute_actionlist($ref2->{"Action"}, $desc);

            }

          } 

          elsif ($below_threshold) {

            # if we are already outside time window but still below
            # threshold, slide the window forward

            push @{$ref2->{"Times"}}, $time;

            update_times($ref2, $time);

          }

        # ----- if we are already checking 2nd threshold...

        } else {

          # inside_window - TRUE if we are still in time window
          # below_threshold - TRUE if we were below threshold before this event

          $inside_window = ($time - $ref2->{"Time"} <= $ref->{"Window2"});
          $below_threshold = (scalar(@{$ref2->{"Times"}}) < $ref->{"Threshold2"});


          if ($inside_window  &&  $below_threshold) {

            # if we are both inside time window and below threshold,
            # we can increase the counter (this threshold is considered
            # as crossed if counter > threshold, counter == threshold
            # is still permitted

            push @{$ref2->{"Times"}}, $time;

          }

          elsif ($inside_window) {

            # if we are inside the time window and below_threshold == FALSE
            # then together with current event we have crossed the threshold
            # (counter > threshold). So we have to slide the window.

            if ($ref->{"Threshold2"}) {

              shift @{$ref2->{"Times"}};
              push @{$ref2->{"Times"}}, $time;

              $ref2->{"Time"} = $ref2->{"Times"}->[0];

            } else { $ref2->{"Time"} = $time; }

          } 

          else {

            # if we have reached here, we must be outside time window
            # and also below threshold, since threshold crossing would
            # have already been detected by previous code block.
            # So we can execute an action.

            execute_actionlist($ref2->{"Action2"}, $ref2->{"Desc2"});

            # since action was just executed we can consider previous
            # application of this rule finished. So this event is again 
            # first one and rule will be started again

            copy_actionlist($ref->{"Action"}, \@action);
            copy_actionlist($ref->{"Action2"}, \@action2);

            $desc2 = $ref->{"Desc2"};

            if (scalar(@subst)) { 

              subst_actionlist(\@subst, \@action, '$');
              subst_string(\@subst, $desc2, '$'); 
              subst_actionlist(\@subst, \@action2, '$');

            }

            $corr_list{$key} = { "Time" => $time, 
                                 "Type" => $ref->{"Type"},
                                 "File" => $conffile,
                                 "ID" => $ref->{"ID"},
                                 "Times" => [ $time ], 
                                 "Window" => $ref->{"Window"}, 
                                 "Context" => [ @context ],
                                 "Desc" => $desc,
                                 "Action" => [ @action ],
                                 "Threshold" => $ref->{"Threshold"}, 
                                 "2ndPass" => 0,
                                 "Window2" => $ref->{"Window2"}, 
                                 "Threshold2" => $ref->{"Threshold2"}, 
                                 "Desc2" => $desc2,
                                 "Action2" => [ @action2 ] };

            if ($ref->{"Threshold"} == 1) {

              $corr_list{$key}->{"2ndPass"} = 1;
              $corr_list{$key}->{"Times"} = [];

              execute_actionlist(\@action, $desc);

            }

          }

        }

      }

      # ------------------------------------------------------------
      # SUPPRESS rule
      # ------------------------------------------------------------

      elsif ($ref->{"Type"} == $SUPPRESS)  { return 1; }

      # ------------------------------------------------------------

      # if match was found and rule's continue-parameter
      # is set to DontCont, return 1, otherwise return 0

      if ($ref->{"WhatNext"} == $DONTCONT)  { return 1; }

    } else {

      # if match was not found and rule is of type Pair*, look also for 
      # all active correlation operations associated with the current 
      # rule and check if 2nd pattern matches

      if ( ($ref->{"Type"} == $PAIR  ||  $ref->{"Type"} == $PAIR_W_WINDOW)  
           &&  scalar(%{$ref->{"Events"}}) ) {

        if (process_rules2($ref)  &&  
            $ref->{"WhatNext2"} == $DONTCONT)  { return 1; }

      }

    }

  }

  return 0;

}



# Parameters: par1 - reference to a rule
# Action: search the events associated with the rule and check, 
#         if there is a matching event for the current content of 
#         input buffer. If there were 1 or more matches found, 
#         return 1, otherwise return 0

sub process_rules2 {

  my($elem) = $_[0];
  my($key, $ref, $ret);
  my($match_found, @subst);
  my($type, $window);
  my($pattype2, $patlines2, $desc2);
  my(@context2, @action2);


  $ret = 0;   # shows if matches were found

  $type = $elem->{"Type"};
  $pattype2 = $elem->{"PatType2"};
  $patlines2 = $elem->{"PatLines2"};
  $window = $elem->{"Window"};

  while ( ($key, $ref) = each(%{$elem->{"Events"}}) ) {

    # Check if last N lines of input buffer match the pattern
    # If match was found, set $match_found to 1
    # If there were any backreferences, assign them to @subst,
    # otherwise leave @subst empty

    if ($pattype2 == $REGEXP) {

      $match_found = 
        match_regexp($patlines2, $ref->{"Pattern2"}, \@subst);

    } elsif ($pattype2 == $SUBSTR) {

      $match_found = match_substr($patlines2, $ref->{"Pattern2"});
      @subst = ();

    }

    if ($match_found) {

      # Check if rule context is valid at the moment

      if (scalar(@{$ref->{"Context2"}})) {

        copy_context($ref->{"Context2"}, \@context2);

        if (scalar(@subst))  { subst_context(\@subst, \@context2, '$'); }

        if (!valid_formula(\@context2))  { next; }  

      } else { @context2 = (); }


      # set $ret to 1 since match has been found

      $ret = 1;

      # increment the counter that shows the usage of the rule
      # (just for statistical purposes)

      ++$elem->{"MatchCount"};


      # --- PAIR rule

      if ($type == $PAIR) {

        # if we are inside time window, execute 2nd action,
        # otherwise do nothing

        if (!$window  ||  time() - $ref->{"Time"} <= $window) {

          copy_actionlist($ref->{"Action2"}, \@action2);

          $desc2 = $ref->{"Desc2"};

          if (scalar(@subst)) { 

            subst_string(\@subst, $desc2, '$'); 
            subst_actionlist(\@subst, \@action2, '$');

          }

          execute_actionlist(\@action2, $desc2);

        }

        # now we can consider rule application finished, and list
        # elements concerning the event can be deleted 

        delete $corr_list{$key};
        delete $elem->{"Events"}->{$key};

      }

      # --- PAIR_W_WINDOW rule

      elsif ($type == $PAIR_W_WINDOW) {

        # if we are inside time window, execute 2nd action,
        # otherwise execute 1st action

        if (time() - $ref->{"Time"} <= $window) {

          copy_actionlist($ref->{"Action2"}, \@action2);

          $desc2 = $ref->{"Desc2"};

          if (scalar(@subst)) { 

            subst_string(\@subst, $desc2, '$'); 
            subst_actionlist(\@subst, \@action2, '$');

          }

          execute_actionlist(\@action2, $desc2);

        } else {

          execute_actionlist($ref->{"Action"}, $ref->{"Desc"});

        }

        # now we can consider rule application finished, and list
        # elements concerning the event can be deleted 

        delete $corr_list{$key};
        delete $elem->{"Events"}->{$key};

      }

    }
    
  }

  # if there were 1 or more matches found, return 1, otherwise return 0

  return $ret;

}



# Parameters: -
# Action: search lists %corr_list, %context_list, @calendar and 
#         @pending_events, performing timed tasks that are associated 
#         with elements and removing obsolete elements

sub process_lists {

  my($key, $ref, $config);
  my($time, $diff, @time);
  my($event, @buffer, $lastdayofmonth);
  my($minute, $hour, $day, $month, $weekday);


  # remove obsolete elements from %context_list

  while (($key, $ref) = each(%context_list)) {

    if ( $ref->{"Window"}  &&  
         time() - $ref->{"Time"} > $ref->{"Window"}  &&
         !exists($ref->{"DeleteInProgress"}) ) {

      log_msg("Deleting stale context '$key'...");

      if (scalar(@{$ref->{"Action"}})) { 

        $ref->{"DeleteInProgress"} = 1;
        execute_actionlist($ref->{"Action"}, $ref->{"Desc"}); 

      }

      delete $context_list{$key};

    }

  }

  # move pending events that have become relevant from 
  # @pending_events list to @events list

  if (scalar(@pending_events)) {

    @buffer = ();

    foreach $ref (@pending_events) {

      if (time() >= $ref->[0]) {

        $event = $ref->[1];
        log_msg("Creating event '$event'...");
        push @events, $event;

      } else { push @buffer, $ref; }

    }  

    @pending_events = @buffer;

  }

  # process CALENDAR rules

  @time = localtime(time());
  $minute = $time[1];
  $hour = $time[2];
  $day = $time[3];
  $month = $time[4];
  $weekday = $time[6];

  $lastdayofmonth = ((localtime(time()+86400))[3] == 1);

  foreach $ref (@calendar) {

    if (scalar(@{$ref->{"Context"}})) {

      if (!valid_formula($ref->{"Context"}))  { next; }  

    }

    # if we have already performed this task in current minute, skip

    if ($minute == $ref->{"LastMinute"} && 
        $hour == $ref->{"LastHour"} &&
        $day == $ref->{"LastDay"} && 
        $month == $ref->{"LastMonth"} &&
        $weekday == $ref->{"LastWeekday"})  { next; }

    # if one (or more) of the conditions does not hold, skip

    if (!exists($ref->{"Minutes"}->{$minute}))  { next; }
    if (!exists($ref->{"Hours"}->{$hour}))  { next; }
 
    if ( !exists($ref->{"Days"}->{$day}) &&
         !($lastdayofmonth && exists($ref->{"Days"}->{"0"}))
       )  { next; } 

    if (!exists($ref->{"Months"}->{$month}))  { next; }
    if (!exists($ref->{"Weekdays"}->{$weekday}))  { next; }

    # execute the action list of the calendar event 
    # and save current time

    execute_actionlist($ref->{"Action"}, $ref->{"Desc"});

    $ref->{"LastMinute"} = $minute;
    $ref->{"LastHour"} = $hour;
    $ref->{"LastDay"} = $day;
    $ref->{"LastMonth"} = $month;
    $ref->{"LastWeekday"} = $weekday;

    ++$ref->{"MatchCount"};

  }

  # perform timed tasks that are associated with elements of
  # %corr_list and remove obsolete elements

  while (($key, $ref) = each(%corr_list)) {

    $time = time();
    $diff = $time - $ref->{"Time"};
    $config = $configuration{$ref->{"File"}}->[$ref->{"ID"}];

    # ------------------------------------------------------------ 
    # SINGLE_W_SUPPRESS rule
    # ------------------------------------------------------------ 

    if ($ref->{"Type"} == $SINGLE_W_SUPPRESS) {

      # if we are outside time window, list element is obsolete
      # and can be removed 

      if ($diff > $ref->{"Window"})  { delete $corr_list{$key}; }

    }

    # ------------------------------------------------------------ 
    # PAIR rule
    # ------------------------------------------------------------ 

    elsif ($ref->{"Type"} == $PAIR) {

      # if we are outside time window, list elements are obsolete
      # and can be removed 

      if ($ref->{"Window"}  &&  $diff > $ref->{"Window"}) {

        delete $config->{"Events"}->{$key};
        delete $corr_list{$key};
      
      }

    }

    # ------------------------------------------------------------ 
    # PAIR_W_WINDOW rule
    # ------------------------------------------------------------ 

    elsif ($ref->{"Type"} == $PAIR_W_WINDOW) {

      # if we are outside time window, 1st action must be executed.
      # After that rule application is finished, so list elements are 
      # obsolete and can be removed 

      if ($diff > $ref->{"Window"}) {

        execute_actionlist($ref->{"Action"}, $ref->{"Desc"});

        # check if the rule has already cleaned itself up 
        # (i.e. 'reset 0' is part of the action list that was just executed)
 
        if (exists($corr_list{$key})) 
            { delete $config->{"Events"}->{$key}; }

        delete $corr_list{$key};

      }

    }

    # ------------------------------------------------------------ 
    # SINGLE_W_THRESHOLD rule
    # ------------------------------------------------------------ 

    elsif ($ref->{"Type"} == $SINGLE_W_THRESHOLD) {

      if ($diff > $ref->{"Window"}) {

        if (scalar(@{$ref->{"Times"}}) < $ref->{"Threshold"}) {

          # If we are outside time window and threshold is not exceeded, 
          # try to slide the window. If all events are gone after sliding,
          # remove the list element as obsolete.

          update_times($ref, $time);

          if (!scalar(@{$ref->{"Times"}}))  { delete $corr_list{$key}; }

        } else {

          # If we are outside time window and threshold is exceeded, 
          # remove the list element as obsolete.

          delete $corr_list{$key};

        }

      }

    }

    # ------------------------------------------------------------ 
    # SINGLE_W_2_THRESHOLDS rule
    # ------------------------------------------------------------ 

    elsif ($ref->{"Type"} == $SINGLE_W_2_THRESHOLDS) {

      if (!$ref->{"2ndPass"}) {

        # If we are outside 1st time window, try to slide the window.
        # If all events are gone after sliding, remove the list element 
        # as obsolete

        if ($diff > $ref->{"Window"}) {

          update_times($ref, $time);

          if (!scalar(@{$ref->{"Times"}}))  { delete $corr_list{$key}; }

        }

      } else {

        # If we are outside 2nd time window and list element
        # has not been removed, we can conclude that 2nd threshold was
        # not exceeded, and so 2nd action can be executed.
        # After that rule application is finished and list element
        # is removed as obsolete

        if ($diff > $ref->{"Window2"}) {

          execute_actionlist($ref->{"Action2"}, $ref->{"Desc2"});

          delete $corr_list{$key};

        }

      }

    }


  }

}



#################################################
# Functions related to reporting and data dumping
#################################################


# Parameters: par1 - reference to a action list
# Action: convert action list to a string representation

sub actionlist2str {

  my($actionlist) = $_[0];
  my($i, $j);
  my($result);


  $i = 0;
  $j = scalar(@{$actionlist});
  $result = "";

  while ($i < $j) {

    if ($actionlist->[$i] == $NONE) { 

      $result .= "none"; 
      ++$i;

    }

    elsif ($actionlist->[$i] == $LOGONLY) { 

      $result .= "logonly"; 
      ++$i;

    } 

    elsif ($actionlist->[$i] == $SHELLCOMMAND) { 

      $result .= "shellcmd " . $actionlist->[$i+1]; 
      $i += 2;

    } 

    elsif ($actionlist->[$i] == $PIPE) {

      $result .= "pipe " . $actionlist->[$i+1] . " " . $actionlist->[$i+2];
      $i += 3;
 
    }

    elsif ($actionlist->[$i] == $CREATECONTEXT) { 

      $result .= "create " . $actionlist->[$i+1] . " " . $actionlist->[$i+2];

      if (scalar(@{$actionlist->[$i+3]})) {

        $result .= " (" . actionlist2str($actionlist->[$i+3]) . ")";

      }

      $i += 4; 

    } 

    elsif ($actionlist->[$i] == $DELETECONTEXT) { 

      $result .= "delete " . $actionlist->[$i+1]; 
      $i += 2;

    } 

    elsif ($actionlist->[$i] == $SETCONTEXT) {
 
      $result .= "set " . $actionlist->[$i+1] . " " . $actionlist->[$i+2];

      if (scalar(@{$actionlist->[$i+3]})) {

        $result .= " (" . actionlist2str($actionlist->[$i+3]) . ")";

      }

      $i += 4;
 
    }

    elsif ($actionlist->[$i] == $ADD) { 

      $result .= "add " . $actionlist->[$i+1] . " " . $actionlist->[$i+2]; 
      $i += 3;

    }

    elsif ($actionlist->[$i] == $REPORT) { 

      $result .= "report " . $actionlist->[$i+1] . " " . $actionlist->[$i+2]; 
      $i += 3;

    }

    elsif ($actionlist->[$i] == $EVENT) { 

      $result .= "event " . $actionlist->[$i+1] . " " . $actionlist->[$i+2]; 
      $i += 3;

    }

    elsif ($actionlist->[$i] == $RESET) { 

      $result .= "reset " . $actionlist->[$i+2] . " " . $actionlist->[$i+3]; 
      $i += 4;

    }

    else { $result .= "unknown action type"; }

    $result .= "; ";

  }

  return $result;

}



# Parameters: par1 - pattern type
#             par2 - pattern lines
#             par3 - pattern
# Action: convert pattern to a printable representation

sub pattern2str {

  my($type) = $_[0];
  my($lines) = $_[1];
  my($pattern) = $_[2];
  

  if ($type == $SUBSTR) { 
    return "substring for $lines line(s): $pattern"; 
  } 

  elsif ($type == $REGEXP) {
    return "regexp for $lines line(s): $pattern";
  } 

  else { return "Unknown pattern type"; }

}



# Parameters: par1 - reference to a context formula
# Action: convert given context to a printable representation

sub context2str {

  my($ref) = $_[0];
  my($i, $j, $op1, $op2);
  my(@stack, $result);


  $i = 0;
  $j = scalar(@{$ref});
  @stack = ();

  while ($i < $j) {

    if ($ref->[$i] == $EXPRESSION) {

      $op1 = $ref->[$i+1];
      push @stack, "(" . context2str($op1) . ")";

      $i += 2;

    }

    elsif ($ref->[$i] == $OPERAND) {

      $op1 = $ref->[$i+1];
      push @stack, $op1;

      $i += 2;
    }

    elsif ($ref->[$i] == $NEGATION) {

      $op1 = pop @stack;
      push @stack, "!" . $op1;

      ++$i;
    }

    elsif ($ref->[$i] == $AND) {

      $op2 = pop @stack;
      $op1 = pop @stack;

      push @stack, $op1 . " && " . $op2;

      ++$i;
    }

    elsif ($ref->[$i] == $OR) {

      $op2 = pop @stack;
      $op1 = pop @stack;

      push @stack, $op1 . " || " . $op2;

      ++$i;
    }

  }

  $result = pop @stack;

  if (!defined($result))  { $result = ""; }

  return $result;

}



# Parameters: par1 - filehandle
#             par2 - list element key
#             par3 - reference to list element
# Action: print given list element to the filehandle

sub print_element {

  my($handle) = $_[0];
  my($key) = $_[1];
  my($ref) = $_[2];
  my($conffile, $id, $time);


  print $handle "Key:\t\t\t\t", $key, "\n";
  print $handle "Start of correlation operation:\t", 
                scalar(localtime($ref->{"Time"})), "\n";

  $conffile = $ref->{"File"};
  $id = $ref->{"ID"};
  $config = $configuration{$conffile}->[$id];

  print $handle "Configuration file:\t\t", $conffile, "\n";
  print $handle "Rule number:\t\t\t", $id+1, "\n";
  print $handle "Rule internal ID:\t\t", $id, "\n";

  if ($ref->{"Type"} == $SINGLE_W_SUPPRESS) {

    print $handle "Type:\t\t\t\t";
    print $handle "SingleWithSuppress\n";

    if ($config->{"WhatNext"} == $DONTCONT) {
      print $handle "Behaviour after match:\t\t", "don't continue\n";
    } else {
      print $handle "Behaviour after match:\t\t", "take next\n";
    }

    print $handle "Pattern:\t\t\t";
    print $handle pattern2str($config->{"PatType"},
                  $config->{"PatLines"}, $config->{"Pattern"});
    print $handle "\n";

    print $handle "Context:\t\t\t";
    print $handle context2str($ref->{"Context"});
    print $handle "\n";

    print $handle "Event:\t\t\t\t", $ref->{"Desc"}, "\n";

    print $handle "Action:\t\t\t\t";
    print $handle actionlist2str($ref->{"Action"});
    print $handle "\n";

    print $handle "Window:\t\t\t\t", $ref->{"Window"}, " seconds\n";

    print $handle "\n";

  }

  elsif ($ref->{"Type"} == $PAIR) {

    print $handle "Type:\t\t\t\t";
    print $handle "Pair\n";

    if ($config->{"WhatNext"} == $DONTCONT) {
      print $handle "Behaviour after 1st match:\t", "don't continue\n";
    } else {
      print $handle "Behaviour after 1st match:\t", "take next\n";
    }

    print $handle "1st Pattern:\t\t\t";
    print $handle pattern2str($config->{"PatType"},
                  $config->{"PatLines"}, $config->{"Pattern"});
    print $handle "\n";

    print $handle "1st Context:\t\t\t";
    print $handle context2str($ref->{"Context"});
    print $handle "\n";

    print $handle "1st Event:\t\t\t", $ref->{"Desc"}, "\n";

    print $handle "1st Action:\t\t\t";
    print $handle actionlist2str($ref->{"Action"});
    print $handle "\n";

    if ($config->{"WhatNext2"} == $DONTCONT) {
      print $handle "Behaviour after 2nd match:\t", "don't continue\n";
    } else {
      print $handle "Behaviour after 2nd match:\t", "take next\n";
    }

    print $handle "2nd Pattern:\t\t\t";
    print $handle pattern2str($config->{"PatType2"},
                  $config->{"PatLines2"}, $ref->{"Pattern2"});
    print $handle "\n";

    print $handle "2nd Context:\t\t\t";
    print $handle context2str($ref->{"Context2"});
    print $handle "\n";

    print $handle "2nd Event:\t\t\t", $ref->{"Desc2"}, "\n";

    print $handle "2nd Action:\t\t\t";
    print $handle actionlist2str($ref->{"Action2"});
    print $handle "\n";

    if ($ref->{"Window"}) {
      print $handle "Window:\t\t\t\t", $ref->{"Window"}, " seconds\n";
    } else {
      print $handle "Window:\t\t\t\t", "infinite\n";
    }

    print $handle "\n";

  }

  elsif ($ref->{"Type"} == $PAIR_W_WINDOW) {

    print $handle "Type:\t\t\t\t";
    print $handle "PairWithWindow\n";

    if ($config->{"WhatNext"} == $DONTCONT) {
      print $handle "Behaviour after 1st match:\t", "don't continue\n";
    } else {
      print $handle "Behaviour after 1st match:\t", "take next\n";
    }

    print $handle "1st Pattern:\t\t\t";
    print $handle pattern2str($config->{"PatType"},
                  $config->{"PatLines"}, $config->{"Pattern"});
    print $handle "\n";

    print $handle "Context:\t\t\t";
    print $handle context2str($ref->{"Context"});
    print $handle "\n";

    print $handle "1st Event:\t\t\t", $ref->{"Desc"}, "\n";

    print $handle "1st Action:\t\t\t";
    print $handle actionlist2str($ref->{"Action"});
    print $handle "\n";

    if ($config->{"WhatNext2"} == $DONTCONT) {
      print $handle "Behaviour after 2nd match:\t", "don't continue\n";
    } else {
      print $handle "Behaviour after 2nd match:\t", "take next\n";
    }

    print $handle "2nd Pattern:\t\t\t";
    print $handle pattern2str($config->{"PatType2"},
                  $config->{"PatLines2"}, $ref->{"Pattern2"});
    print $handle "\n";

    print $handle "2nd Context:\t\t\t";
    print $handle context2str($ref->{"Context2"});
    print $handle "\n";

    print $handle "2nd Event:\t\t\t", $ref->{"Desc2"}, "\n";

    print $handle "2nd Action:\t\t\t";
    print $handle actionlist2str($ref->{"Action2"});
    print $handle "\n";

    print $handle "Window:\t\t\t\t", $ref->{"Window"}, " seconds\n";

    print $handle "\n";

  }

  elsif ($ref->{"Type"} == $SINGLE_W_THRESHOLD) {

    print $handle "Type:\t\t\t\t";
    print $handle "SingleWithThreshold\n";

    if ($config->{"WhatNext"} == $DONTCONT) {
      print $handle "Behaviour after match:\t\t", "don't continue\n";
    } else {
      print $handle "Behaviour after match:\t\t", "take next\n";
    }

    print $handle "Pattern:\t\t\t";
    print $handle pattern2str($config->{"PatType"},
                  $config->{"PatLines"}, $config->{"Pattern"});
    print $handle "\n";

    print $handle "Context:\t\t\t";
    print $handle context2str($ref->{"Context"});
    print $handle "\n";

    print $handle "Event:\t\t\t\t", $ref->{"Desc"}, "\n";

    print $handle "Action:\t\t\t\t";
    print $handle actionlist2str($ref->{"Action"});
    print $handle "\n";

    print $handle "Window:\t\t\t\t", $ref->{"Window"}, " seconds\n";

    print $handle "Threshold:\t\t\t", $ref->{"Threshold"}, "\n";

    print $handle scalar(@{$ref->{"Times"}}), " events observed at:\n";

    foreach $time (@{$ref->{"Times"}}) 
        { print $handle scalar(localtime($time)), "\n"; }

    print $handle "\n";

  }

  elsif ($ref->{"Type"} == $SINGLE_W_2_THRESHOLDS) {

    print $handle "Type:\t\t\t\t";
    print $handle "SingleWith2Thresholds\n";

    if ($config->{"WhatNext"} == $DONTCONT) {
      print $handle "Behaviour after match:\t\t", "don't continue\n";
    } else {
      print $handle "Behaviour after match:\t\t", "take next\n";
    }

    print $handle "Pattern:\t\t\t";
    print $handle pattern2str($config->{"PatType"},
                  $config->{"PatLines"}, $config->{"Pattern"});
    print $handle "\n";

    print $handle "Context:\t\t\t";
    print $handle context2str($ref->{"Context"});
    print $handle "\n";

    print $handle "1st Event:\t\t\t", $ref->{"Desc"}, "\n";

    print $handle "1st Action:\t\t\t";
    print $handle actionlist2str($ref->{"Action"});
    print $handle "\n";

    print $handle "1st Window:\t\t\t", $ref->{"Window"}, " seconds\n";

    print $handle "1st Threshold:\t\t\t", $ref->{"Threshold"}, "\n";

    print $handle "2nd Event:\t\t\t", $ref->{"Desc2"}, "\n";

    print $handle "2nd Action:\t\t\t";
    print $handle actionlist2str($ref->{"Action2"});
    print $handle "\n";

    print $handle "2nd Window:\t\t\t", $ref->{"Window2"}, " seconds\n";

    print $handle "2nd Threshold:\t\t\t", $ref->{"Threshold2"}, "\n";

    print $handle scalar(@{$ref->{"Times"}}), " events observed at ";

    if ($ref->{"2ndPass"}) { 
      print $handle "(checking 2nd threshold):\n"; 
    } else { 
      print $handle "(checking 1st threshold):\n"; 
    }

    foreach $time (@{$ref->{"Times"}})
        { print $handle scalar(localtime($time)), "\n"; }

    print $handle "\n";

  }

}



# Parameters: -
# Action: save some information about the current state of the program
#         to dump file.

sub dump_data {

  my($i, $line, $key, $ref, $conffile, $event);
  my($user, $system, $cuser, $csystem);


  if (open(DUMPFILE, ">$dumpfile")) {

    # print performance statistics

    print DUMPFILE "Performance statistics:\n";
    print DUMPFILE '=' x 60, "\n";

    ($user, $system, $cuser, $csystem) = times();

    print DUMPFILE "Run time: ", time() - $startuptime, " seconds\n";
    print DUMPFILE "User time: $user seconds\n";
    print DUMPFILE "System time: $system seconds\n";
    print DUMPFILE "Child user time: $cuser seconds\n";
    print DUMPFILE "Child system time: $csystem seconds\n";
    print DUMPFILE "Processed input lines: $processedlines\n";

    print DUMPFILE "\n";


    # print rule usage statistics

    print DUMPFILE "Rule usage statistics:\n";
    print DUMPFILE '=' x 60, "\n";

    foreach $conffile (@conffiles) {

      print DUMPFILE "\nStatistics for the rules from $conffile\n";
      print DUMPFILE '-' x 60, "\n";
      $i = 1;

      foreach $ref (@{$configuration{$conffile}}) {

        print DUMPFILE "Rule $i at line ", $ref->{"LineNo"}, 
          " (", $ref->{"Desc"}, ") has matched ", 
            $ref->{"MatchCount"}, " events\n";

        ++$i;

      }

    }

    print DUMPFILE "\n";


    # print content of input buffer

    print DUMPFILE "Content of input buffer:\n";
    print DUMPFILE '-' x 60, "\n";

    for ($i = $bufpos - $bufsize + 1; $i <= $bufpos; ++$i) {

      print DUMPFILE $input_buffer[$i % $bufsize], "\n";

    }

    print DUMPFILE '-' x 60, "\n";
    print DUMPFILE "\n";


    # print content of pending event buffer

    print DUMPFILE "Pending events that are generated by rules:\n";
    print DUMPFILE '-' x 60, "\n";

    foreach $ref (@pending_events) { 

      print DUMPFILE "Event: ", $ref->[1], "\n";
      print DUMPFILE "Will be created at: ", 
                     scalar(localtime($ref->[0])), "\n";
      print DUMPFILE "\n";

    }

    print DUMPFILE '-' x 60, "\n";
    print DUMPFILE "\n";


    # print the list of active event correlation operations

    $i = 0;
    print DUMPFILE "List of active event correlation operations:\n";
    print DUMPFILE '=' x 60, "\n";

    while (($key, $ref) = each(%corr_list)) { 

      print_element(*DUMPFILE, $key, $ref);
      print DUMPFILE '-' x 60, "\n";

      ++$i; 

    }

    print DUMPFILE "Total: $i elements\n\n";


    # print the list of active contexts

    $i = 0;
    print DUMPFILE "List of active contexts:\n";
    print DUMPFILE '=' x 60, "\n";

    while (($key, $ref) = each(%context_list)) { 

      print DUMPFILE "Context Name: ", $key, "\n";
      print DUMPFILE "Creation Time: ", 
                     scalar(localtime($ref->{"Time"})), "\n";

      if ($ref->{"Window"}) {
        print DUMPFILE "Lifetime: ", $ref->{"Window"}, " seconds\n";
      } else {
        print DUMPFILE "Lifetime: infinite\n";
      }

      if (scalar(@{$ref->{"Action"}})) {
        print DUMPFILE "Action on delete: ", 
                       actionlist2str($ref->{"Action"});
        print DUMPFILE " (%s = ", $ref->{"Desc"}, ")\n";
      }

      if (scalar(@{$ref->{"Buffer"}})) {

        print DUMPFILE scalar(@{$ref->{"Buffer"}}), 
                       " events associated with context:\n";

        foreach $event (@{$ref->{"Buffer"}}) 
                { print DUMPFILE $event, "\n"; }

      }

      print DUMPFILE '-' x 60, "\n";
      ++$i;

    }
    
    print DUMPFILE "Total: $i elements\n\n";


    close(DUMPFILE);

  } else {

    log_msg("Cannot open dumpfile $dumpfile!");

  }

}



#################################################################
# Functions related to timing, input handling and signal handling
#################################################################


# Parameters: -
# Action: perform timed tasks and process arrived signals

sub timed_tasks {

  my($histmsg, $pid, $exitcode);


  # if we have waited for new bytes for more than $reopen_timeout 
  # seconds, reopen input stream

  if (defined($reopen_timeout)  &&
      time() - $lastreopen >= $reopen_timeout) {

    log_msg("Timeout reached when reading from $inputfile, reopening!");

    open_input($inputfile, -1);
    $lastreopen = time();

  }

  # if we have waited for new bytes for more than $input_timeout 
  # seconds, execute external script $timeout_script with commandline
  # parameter "1"

  if (defined($input_timeout)  && 
      time() - $lastread >= $input_timeout  &&  !$script_executed) {

    log_msg("No input, executing script $timeout_script 1");

    shell_cmd("$timeout_script %s", "1"); 
    $script_executed = 1;

  }

  # search all lists, performing timed tasks associated with elements
  # and removing obsolete elements

  if (time() - $lastcleanuptime >= $cleantime) {

    process_lists();
    $lastcleanuptime = time();

  }

  # get the exit status of every terminated child process. If a child 
  # was executed as part of SINGLE_W_SCRIPT rule, execute the action 
  # lists if the exit code of terminated child is 0

  for (;;) {

    # get the exit status of next terminated child process and
    # quit the loop if there are no more deceased children

    $pid = waitpid(-1, &WNOHANG);
    if ($pid < 1) { last; }

    # if the child process has really exited (and not stopped), check
    # the exit code of the child and log an error message if the code is
    # not equal to 0. If the child's exit code equals to 0 and it was 
    # started as a part of SINGLE_W_SCRIPT rule, execute the action list

    if (WIFEXITED($?)) {

      $exitcode = $? >> 8;

      if ($exitcode != 0) {

        log_msg("Child $pid terminated with non-zero exitcode $exitcode");

      } elsif (exists($children{$pid})) {

        execute_actionlist($children{$pid}->{"Action"},
                           $children{$pid}->{"Desc"});
      }

      delete $children{$pid};

    }

  }

  # if SIGHUP has arrived, reopen input file and logfile, re-read 
  # configuration and empty all lists concerning events and correlation 
  # information; if SIGABRT has arrived, behave like SIGHUP but preserve 
  # contexts that are active

  if ($refresh || $softrefresh) {

    if ($refresh) {
      log_msg("SIGHUP received: reopening input and logfile, reading configuration and cleaning all internal data structures...");
    }

    elsif ($softrefresh) {
      log_msg("SIGABRT received: reopening input and logfile, reading configuration and cleaning all internal data structures (except contexts)...");
    }

    %corr_list = ();
    %children = ();

    if ($refresh)  { %context_list = (); }

    @events = ();
    @pending_events = ();

    open_input($inputfile, -1);
    $lastreopen = time();

    read_config();

    # reopen the logfile, write the content of temporary buffer
    # to a logfile, and set flags back to zero

    open_logfile($logfile);
    
    if (scalar(@logmsgbuffer)) {

      foreach $histmsg (@logmsgbuffer)  { print LOGFILE $histmsg; }
      @logmsgbuffer = ();

    }

    $refresh = 0;
    $softrefresh = 0;

  }

  # if SIGUSR1 has arrived, create dump file

  if ($dumpdata) {

    log_msg("SIGUSR1 received: dumping data to $dumpfile...");

    dump_data();

    $dumpdata = 0;

  }

  # if SIGUSR2 has arrived, reopen logfile

  if ($openlog) {

    log_msg("SIGUSR2 received: reopening logfile...");

    # reopen the logfile, write the content of temporary buffer
    # to a logfile, and set flags back to zero

    open_logfile($logfile);

    if (scalar(@logmsgbuffer)) {

      foreach $histmsg (@logmsgbuffer)  { print LOGFILE $histmsg; }
      @logmsgbuffer = ();

    }
 
    $openlog = 0;

  }

}



# Parameters: par1 - name of the input file
#             par2 - file position
# Action: Input file will be opened (in non-blocking mode if necessary)
#         and file position will be moved to position par2 
#         (-1 means "seek EOF" and 0 means "don't seek at all").
#         Filehandle of opened input stream will be saved to global
#         filehandle INPUT.

sub open_input {

  my($file) = $_[0];
  my($fpos) = $_[1];
  my($flags);


  # if input is stdin, duplicate it

  if ($file eq "-") {

    if (!open(INPUT, "<&STDIN")) {

      log_msg("Can't dup stdin ($!), exiting!");
      exit(1);

    }

  }

  # if input file is a regular file, open it for reading

  elsif (-f $file) {

    if (!sysopen(INPUT, $file, O_RDONLY)) {

      log_msg("Can't open file $file ($!), exiting!");
      exit(1);

    }

  }

  # if input is a named pipe, open it both for reading and writing
  # (the open would block if there are no writers at the moment,
  # so process pretends to be a writer)

  elsif (-p $file) {

    if (!sysopen(INPUT, $file, O_RDWR)) {

      log_msg("Can't open file $file ($!), exiting!");
      exit(1);

    }

  }

  # unsupported or non-existing input file

  else {

    log_msg("Input file $file is not stdin, regular file, or named pipe!");
    exit(1);

  }

  # if INPUT filehandle is not connected to a regular file,
  # switch INPUT filehandle to a nonblocking mode

  if (! -f INPUT) {

    $flags = 0;
    fcntl(INPUT, F_GETFL, $flags);
    $flags |= O_NONBLOCK;

    if (!fcntl(INPUT, F_SETFL, $flags)) {

      log_msg("Can't switch $file to a non-blocking mode ($!), exiting!");
      exit(1);

    }

  }

  # if INPUT filehandle is connected to a regular file
  # and $fpos == -1 or $fpos > 0, seek the given position in the file

  if (-f INPUT) {

    if ($fpos == -1  &&  !sysseek(INPUT, 0, SEEK_END)) {
      log_msg("Can't seek EOF in file $file ($!)!");
    }
    elsif ($fpos > 0  &&  !sysseek(INPUT, $fpos, SEEK_SET)) {
      log_msg("Can't seek position $fpos in file $file ($!)!");
    }

  }

}



# Parameters: par1 - name of the input file
# Action: check if input file has been removed, recreated or truncated.
#         Return 1 if input file has changed and should be reopened; 
#         return 0 if the file has not changed or should not be
#         reopened right now

sub input_shuffled {

  my($file) = $_[0];
  my(@oldstat, @newstat);


  # standard input is always intact (it can't be recreated or truncated)

  if ($file eq "-")  { return 0; }

  # stat the input filehandle and report input as shuffled if stat fails

  @oldstat = stat(INPUT);

  if (!scalar(@oldstat))  { return 1; }

  # stat the input file and return 0 if stat fails (e.g., input file has 
  # been removed and not recreated yet, so we can't reopen it now)

  @newstat = stat($file);

  if (!scalar(@newstat))  { return 0; }

  # compare two stat structures in order to detect input shuffling:
  # check if i-node numbers of filehandle and input file are different,
  # and check if file size has decreased

  if ($oldstat[0] != $newstat[0] || $oldstat[1] != $newstat[1]) { return 1; }

  if (-f INPUT && $oldstat[7] > $newstat[7]) { return 1; }

  return 0;

}



# Parameters: - 
# Action: read next line from input stream. If input stream has no data
#         available, subroutine will wait until fresh data becomes 
#         available and return next line from input stream ('\n' will be
#         removed from the end of line). If there are pending events
#         in event buffer, event buffer takes precedence over input
#         stream and the first event from event buffer will be returned.

sub read_line {

  my($char, $line, $rin);


  for (;;) {

    # if there are pending events in event buffer,
    # return with the first event from the buffer

    if (scalar(@events))  { return shift @events; }

    # if INPUT filehandle is connected to a regular file,
    # try to read new bytes and if no new data is available,
    # sleep for $poll_timeout seconds with the help of select()

    if (-f INPUT) {

      $line = "";

      while (sysread(INPUT, $char, 1)) {

        $line .= $char;
        if ($char eq "\n")  { last; }

      }

      if (!length($line)) { 

        select(undef, undef, undef, $poll_timeout); 
        $line = undef;
        
      }

    }

    # if INPUT filehandle is not connected to a regular file,
    # poll input stream for $poll_timeout seconds with select(),
    # and if there is new data available, try to read it

    else {

      $rin = '';
      vec($rin, fileno(INPUT), 1) = 1;

      if (select($rin, undef, undef, $poll_timeout)) {

        $line = "";

        while (sysread(INPUT, $char, 1)) {

          $line .= $char;
          if ($char eq "\n")  { last; }

        }

        if (!length($line))  { $line = undef; }

        # if select() reported available data but read returned nothing,
        # and the IO error was not caused by a caught signal, exit

        if (!defined($line)  &&  $! != EINTR  &&  $! != EAGAIN) {

          log_msg("Error when reading from $inputfile ($!), exiting!");
          exit(1);

        }

      } else { $line = undef; }

    }
  
    # if read was successful, exit the loop

    if (defined($line))  { last; }


    # if there were no new bytes in input stream and -notail option
    # was specified in commandline, exit

    if (!$tail)  { exit(0); }

    # if there were no new bytes in input stream and input file has 
    # been shuffled, reopen input file and start to process it from 
    # the beginning

    if (input_shuffled($inputfile)) {

      log_msg("Input file $inputfile has been recreated or truncated, reopening!");
      open_input($inputfile, 0);
      $lastreopen = time();

    }

    # Perform timed tasks

    timed_tasks();

  }


  # If "$timeout_script 1" has been executed before,
  # execute it again with commandline parameter "0"

  if ($script_executed) {

    log_msg("Input available again, executing script $timeout_script 0");
    shell_cmd("$timeout_script %s", "0");

    $script_executed = 0;

  }

  # Set the global variables to reflect the time of last successful read

  $lastread = $lastreopen = time();

  chomp($line);
  return $line;

}



# Parameters: -
# Action: on arrival of SIGHUP set flag $refresh

sub hup_handler {

  $SIG{HUP} = \&hup_handler;
  $refresh = 1;

}               



# Parameters: -
# Action: on arrival of SIGABRT set flag $softrefresh

sub abrt_handler {

  $SIG{ABRT} = \&abrt_handler;
  $softrefresh = 1;

}               



# Parameters: -
# Action: on arrival of SIGUSR1 set flag $dumpdata

sub usr1_handler {

  $SIG{USR1} = \&usr1_handler;
  $dumpdata = 1;

}               



# Parameters: -
# Action: on arrival of SIGUSR2 set flag $openlog

sub usr2_handler {

  $SIG{USR2} = \&usr2_handler;
  $openlog = 1;

}               



##################################################################
# ------------------------- MAIN PROGRAM -------------------------
##################################################################

### Open logfile

open_logfile($logfile);


# If -detach flag was specified, chdir to / for not disturbing future 
# unmount of current filesystem. Must be done before read_config() to 
# receive error messages about scripts that would not be found at runtime

if ($detach) { 

  log_msg("Changing working directory to /");
  chdir('/'); 

}


### Read in configuration

read_config();


### Open input stream

if ($tail) { open_input($inputfile, -1); } 
  else { open_input($inputfile, 0); }


### Detach process from controlling tty, if -detach flag was specified

if ($detach) {

  # if stdin was specified as input, but it is still connected to terminal
  # (i.e. not redirected from a file or pipe), we can't detach

  if ($inputfile eq "-"  &&  -t STDIN) {

    log_msg("Can't detach from tty (stdin is specified as input)");
    exit(1);

  }

  # if stdin, stdout and stderr are associated with controlling 
  # terminal, redirect them to /dev/null

  if (-t STDIN)  { open(STDIN, '/dev/null'); }
  if (-t STDOUT)  { open(STDOUT, '/dev/null'); }
  if (-t STDERR)  { open(STDERR, '/dev/null'); }

  if (fork())  { exit 0; }

  POSIX::setsid();

  log_msg("Detached from controlling terminal");

}


### Create pidfile - must be done after detaching

if (open(PIDFILE, ">$pidfile")) {

  print PIDFILE "$$\n";
  close PIDFILE;

} else {

  log_msg("Can't open pidfile $pidfile for writing!");
  exit(1);

}


### Set signal handlers

$refresh = 0;
$SIG{HUP} = \&hup_handler;

$softrefresh = 0;
$SIG{ABRT} = \&abrt_handler;

$dumpdata = 0;
$SIG{USR1} = \&usr1_handler;

$openlog = 0;
$SIG{USR2} = \&usr2_handler;


### Set various global variables

$lastcleanuptime = $lastread = $lastreopen = $startuptime = time();
$script_executed = 0;
$processedlines = 0;


### Initialize input buffer

for ($i = 0; $i < $bufsize; ++$i)  { $input_buffer[$i] = ""; }
$bufpos = -1;


### Initialize correlation list, context list, 
### buffer list and child process list

%corr_list = ();
%context_list = ();
%children = ();


### Initialize event buffers

@events = ();
@pending_events = ();


### Initialize log-message buffer

@logmsgbuffer = ();


### Read lines from input stream and process them

for (;;) {

  # if there are pending events in event buffer, read new line
  # from event buffer, otherwise read new line from input stream.

  $line = read_line();

  # update input buffer

  $bufpos = ($bufpos + 1) % $bufsize;
  $input_buffer[$bufpos] = $line;

  # process rules from configuration files and perform timed tasks

  foreach $conffile (@conffiles)  { process_rules($conffile); }

  timed_tasks();

  ++$processedlines;

}
